// CopyFileThread.cpp
//
// Implementation of CopyFileThread class.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "CopyFileThread.h"

// TTFS
#include "TTFSFileSystem.h"
#include "TTFSFileEngine.h"

// TomTomSupporterBase
#include "IProgressReporter.h"
#include "IPackageValidator.h"

// Qt
#include <QDir>
#include <QFile>
#include <QFileInfo>
#include <QByteArray>


CopyFileThread::CopyFileThread(const QString& aFilePath, const QString& aFolder,
        IProgressReporter* aReporter, IPackageValidator* aValidator,
        QObject* aParent) :
    RegisteredThread(aParent, "CopyFileThread"),
    iFilePath(aFilePath),
    iFolder(aFolder),
    iReporter(aReporter),
    iValidator(aValidator),
    iError(CopyFileThread::_errUNDEF)
{
    ; // empty
}

CopyFileThread::~CopyFileThread()
{
    ; // empty
}

void CopyFileThread::run()
{
    iError = _errUNDEF;
    
    QFile theFile(iFilePath);
    if (!theFile.exists())
    {
        iError = errSourceFileDoesNotExist;
        return;
    }

    QDir theDir(iFolder);
    QFileInfo theFileInfo(iFilePath);
    QString theTargetPath(theDir.filePath(theFileInfo.fileName()));
    QFile theTargetFile(theTargetPath);
    if (theTargetFile.exists())
    {
        iError = errTargetFileAlreadyExists;
        return;
    }
    
    qint64 theFileSize = theFile.size();
    bool theTargetOnEXT3(TTFSFileEngine::canHandle(theTargetPath));
    
    if (theTargetOnEXT3)
    {
        TTFSFileEngine theFileEngine(iFolder);
        bool theSpaceAvailable = (quint64)theFileSize <= theFileEngine.freeBytes();
        if (!theSpaceAvailable)
        {
            iError = errNotEnoughSpaceAvailable;
            return;
        }
    }

    if (iReporter || iValidator)
    {
        if (theFile.open(QIODevice::ReadOnly))
        {
            if (theTargetFile.open(QIODevice::WriteOnly))
            {
                theTargetFile.resize(theFileSize);

                const qint64 the4KBytes = 16 * 1024;
                QByteArray theBuffer;
                theBuffer.resize(the4KBytes);
                qint64 thePosition = 0;
                do 
                {
                    qint64 theBytesToRead = std::min<qint64>(the4KBytes, theFileSize - thePosition);
                    qint64 theBytesRead = theFile.read(theBuffer.data(), theBytesToRead);
                    if (theBytesRead > 0)
                    {
                        theBuffer.resize((int)theBytesToRead);
                        bool theIsValid = !iValidator || iValidator->Validate(theBuffer);
                        
                        if (!theIsValid)
                        {
                            iError = errInvalidData;
                            break;
                        }
                        qint64 theBytesWritten = theTargetFile.write(theBuffer, theBytesRead);
                        if (theBytesWritten < theBytesRead)
                        {
                            iError = errUnableToWriteAllData;
                            break;
                        }
                    }
                    thePosition += theBytesRead;
                    if (iReporter)
                    {
                        iReporter->Report(thePosition, theFileSize);
                    }
                }
                while (thePosition < theFileSize);
                
                theTargetFile.close();
            }
            else
            {
                iError = errUnableToOpenFileForWriting;
            }
            theFile.close();
        }
    }
    else
    {
        theFile.copy(theTargetPath);
    }
}

CopyFileThread::Error CopyFileThread::GetLastError() const
{
    return iError;
}

struct ErrorString
{
    CopyFileThread::Error error;
    const char* str;
};
    
static struct ErrorString sErrorStringList[] =
{
    { CopyFileThread::errSourceFileDoesNotExist,     "Source file does not exist" },
    { CopyFileThread::errTargetFileAlreadyExists,    "Target file already exists" },
    { CopyFileThread::errUnableToWriteAllData,       "Unable to write all the data." },
    { CopyFileThread::errInvalidData,                "Invalid data." },
    { CopyFileThread::errUnableToOpenFileForWriting, "Unable to open file for writing." },
    { CopyFileThread::errNotEnoughSpaceAvailable,    "Not enough space available." },
    { CopyFileThread::_errUNDEF,                     "" } // Must be last
};

/*static*/
QString CopyFileThread::GetErrorString(CopyFileThread::Error aError)
{
    Q_ASSERT((CopyFileThread::_errMIN < aError) && (aError < CopyFileThread::_errMAX));
    
    QString theRetval;
    for (int i = 0; true; ++i)
    {
        if (sErrorStringList[i].error == CopyFileThread::_errUNDEF || 
            sErrorStringList[i].error == aError)
        {
            theRetval = sErrorStringList[i].str;
            break;
        }
    }
    return theRetval;
}
