// GetDeviceSize.cpp
//
// Implementation to get the device size (the size of TTFS filesystem).
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "GetDeviceSize.h"

// Qt
#include <QFileInfo>
#include <QString>

// ext2fs
#include <ext2fs/ext2_fs.h>
#include <ext2fs/ext2fs.h>

extern "C" {

const int cImageFilesMax = 256;

// File size constraint: the file size should be equal 2GB or less for last file.
const qint64 cFileSizeMax = ((qint64)1 << 31);

errcode_t ttfs_get_device_size(const char *file,
                                         int blocksize, 
                                         blk64_t *retblocks)
{
    errcode_t retval = EXT2_ET_BAD_DEVICE_NAME;
    qint64 theTTFSSize = 0;
    QString theFilename;
    int theFileIndex;
    bool theSuccess;

    if (blocksize <= 0)
    {
        return EXT2_ET_UNEXPECTED_BLOCK_SIZE;
    }
    if (file == 0)
    {
        return EXT2_ET_BAD_DEVICE_NAME;
    }
    theFilename = QString(file);
    // at least once it should be executed.
    do
    {
        QFileInfo theFileInfo(theFilename);
        if (theFileInfo.exists() && theFileInfo.isFile())
        {
            retval = 0;
            theTTFSSize += theFileInfo.size();
            // Requirement: each file is 2GB, last one could be less.
            // As the file size is less than 2GB, it means last file.
            if (theFileInfo.size() == cFileSizeMax)
            {
                // Get the extention of a file: it is a number.
                theFileIndex = theFileInfo.suffix().toInt(&theSuccess);
                if (theSuccess)
                {
                    theFilename = theFileInfo.path() + "/" +
                                    theFileInfo.completeBaseName() + "." +
                                    QString("%1").arg(theFileIndex + 1, 3, 10,  QLatin1Char('0'));
                }
                else
                {
                    // Break the loop and return.
                    break;
                }
            }
            else
            {
                // Break the loop and return.
                break;
            }
        }
        else
        {
            // Break the loop and return.
            break;
        }
    }
    while (theFileIndex < cImageFilesMax);

    if (theTTFSSize > 0)
    {
        *retblocks = (blk_t)(theTTFSSize / (qint64)blocksize);
    }
    return retval;
}

} /* extern "C" */
