// TTFSDebug.h
//
// TTFSDebug provides functionality debugging capabilities based on
// QtDebug, for debugging TTFS.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#pragma once

#include <QtCore>
#include <QtDebug>

#include "Logger.h"
#include "TTFSLoggerStream.h"

// ----------------------------------------------------------------------------
// DEBUG SPY MACROS
// ----------------------------------------------------------------------------
// The macros are normally used by the caller.
//
// Use:
//   spyFAIL for failure or severe/very important mesages.
//   spyWARN for warnings or important message.
//   spyINFO for informational messages only.
//
// Examples:
//    spyFAIL("Cannot happend. Name:" << Name << "Date:" << Date.toString());
//    spyWARN("Expected Level 3. Now at Level:" << i);
//    spyINFO("Object status is:" << isOk);
//
// Important: Notice the ";" at the end of the lines. They are
// required.
//
#define spyFAIL(logId, args)    SPY_LEVEL(TTFS::spylevelFAIL, logId, args)
#define spyWARN(args)           SPY_LEVEL(TTFS::spylevelWARN, 0, args)
#define spyINFO(args)           SPY_LEVEL(TTFS::spylevelINFO, 0, args)



// ----------------------------------------------------------------------------
// HELPER TYPES AND FUNCTIONS
// ----------------------------------------------------------------------------
namespace TTFS
{
    // This type defines the spy levels.
    enum SpyLevel           
    {
        _spylevelUNDEF,     // For internal use.
        _spylevelMIN,       // For internal use.
        spylevelOFF,        // No messages.
        spylevelFAIL,       // Failures only.
        spylevelWARN,       // Warnings and failures.
        spylevelINFO,       // Informationals, warnings and failures.
        _spylevelMAX        // For internal use.
    };

    // To get/set the current spy level, use spyLevel() and
    // setSpyLevel().
    SpyLevel spyLevel();
    void SetSpyLevel(SpyLevel level);
};

// ----------------------------------------------------------------------------
// NON-DEBUG MODE
// ----------------------------------------------------------------------------
//
// Do not output spy messages and do not evualate the macro parameter either.

#ifdef QT_NO_DEBUG
#define QT_NO_DEBUG_OUTPUT 1

// Note that the if() below is to make sure the macro compiles
// similarly to the debug version of the macro and it is always
// a statement.
//
#define SPY_LEVEL(level, logId, msg) \
    if (true)   \
    {           \
        ;       \
    }           \
    else        \
    {           \
        ;       \
    }

#else

// ----------------------------------------------------------------------------
// DEBUG MODE
// ----------------------------------------------------------------------------
//
// Provide spy messages, using qDebug(). This outputs spy messages to a console
// for a console application, or to the debugger for MSVC.
//
//
// Try to automatically add the function name to the spy message, by using a
// standard compiler macro __FUNCTION__.
//
#ifdef __FUNCTION__
#define SPY_FUNCTION_NAME  __FUNCTION__         // Try the standard macro.
#else
// TODO: __func__ is part of the C99 standard. It is not a MACRO.
// __function__ something that doesn't exist.
#ifdef __function__
#define SPY_FUNCTION_NAME  __function__         // Or try the C99 macro.
#else
#define SPY_FUNCTION_NAME  __FILE__             // Otherwise show filename.
#endif
#endif
 
#define SPY_LEVEL(level, logId, args)                       \
        if (TTFS::spyLevel() >= level)                      \
        {                                                   \
            QString theString;                              \
            TTFSLoggerStream theStream(&theString);         \
            theStream << args;                              \
            if (level >= TTFS::spylevelINFO)                \
            {                                               \
                Logger::LogInformation(SPY_FUNCTION_NAME,   \
                    theString);                             \
            }                                               \
            else if (level == TTFS::spylevelWARN)           \
            {                                               \
                Logger::LogWarning(SPY_FUNCTION_NAME,       \
                    theString);                             \
            }                                               \
            else                                            \
            {                                               \
                 Logger::LogError(logId,                    \
                     SPY_FUNCTION_NAME,                     \
                     theString);                            \
            }                                               \
        }                                                   \
        else                                                \
        {                                                   \
            ;                                               \
        }                                                   \

#endif
