// TTFSDir.cpp
//
// Implementation of TTFSDir class.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "TTFSDir.h"

// TTFS (this software library)
#include "TTFSDebug.h"
#include "TTFSError.h"
#include "TTFSFileSystem.h"
#include "TTFSFileInfo.h"
#include "TTFSTypes.h"
#include "TTFSNode.h"
#include "TTFSIO.h"

// ext2fs
#include <ext2fs_prefix.h>
#include <ext2fs/ext2fs.h>

// Qt
#include <QtCore>
#include <QString>
#include <QDir>



// -----------------------------------------------------------------------------
// STATIC C CALLBACK FUNCTIONS
// -----------------------------------------------------------------------------

extern "C"
{

static int cbGetFileInfo(ext2_dir_entry* dirEnt, int unused1, int unused2, char* unused3,
    void* privateData)
{
    Q_UNUSED(unused1);
    Q_UNUSED(unused2);
    Q_UNUSED(unused3);
    Q_ASSERT(dirEnt);

    spyINFO((dirEnt ? dirEnt->name : "<unknown directory name>"));
    typedef std::pair<TTFSFileInfoList*, ext2_filsys> PrivateDataType;
    PrivateDataType* thePrivateData = (PrivateDataType*) privateData;

    char nameBuf[EXT2_NAME_LEN + 1];
    ext2_dir_entry_2* dirEnt2 = (ext2_dir_entry_2*) dirEnt;
    int len = std::min<int>(dirEnt2->name_len, EXT2_NAME_LEN);
    strncpy(nameBuf, dirEnt2->name, len);
    nameBuf[len] = 0;
    QString theName = QString::fromUtf8(nameBuf);

    ext2_inode iNodeStruct;
    errcode_t err = ext2fs_read_inode(thePrivateData->second, dirEnt2->inode, &iNodeStruct);
    if (!err)
    {
        TTFS::FileType fileType = TTFS::ftUnknown;
        if (dirEnt2->file_type == EXT2_FT_DIR)
        {
            iNodeStruct.i_mode |= LINUX_S_IFDIR; // It is missing for older types of PNDs like Carminat.
            fileType = TTFS::ftDirectory;
        }
        else if (dirEnt2->file_type == EXT2_FT_REG_FILE)
        {
            iNodeStruct.i_mode |= LINUX_S_IFREG; // It is missing for older types of PNDs like Carminat.
            fileType = TTFS::ftRegularFile;
        }
        else 
        {
            ; //!! TODO What needs to be done here?
            // LINUX_S_IFSOCK
            // LINUX_S_IFLNK
            // LINUX_S_IFBLK
            // LINUX_S_IFCHR
            // LINUX_S_IFIFO
        }
        if (fileType != TTFS::ftUnknown)
        {
            thePrivateData->first->push_back(TTFSFileInfo(theName, fileType,
                dirEnt2->inode, iNodeStruct));
        } 
    }
    return 0;
}

} // extern "C"



// -----------------------------------------------------------------------------
// CTOR/DTOR
// -----------------------------------------------------------------------------

TTFSDir::TTFSDir(const QString& aPath, TTFS::OpenMode openMode) :
    TTFSNode(aPath, TTFS::ftDirectory, openMode)
{
    spyINFO("aPath:" << aPath << "openMode:" << openMode);
    Q_ASSERT(isValid());

    // Check if the inode is really a directory.
    bool isFile(false);
    if (iFS)
    {
        ext2_ino_t inode = inodeNr();
        if (inode)
        {
            errcode_t err = ext2fs_check_directory(iFS, inode);
            if (err == EXT2_ET_NO_DIRECTORY)
            {
                spyWARN(aPath << "should be a directory, but it is not");
                isFile = true;
            }
        }
    }

    if (isFile)
    {
        // Clean up. Even though this is programming bug by the caller,
        // we should not continue just issuing an assertion.
        Q_ASSERT(false);
        unmount();
    }
    Q_ASSERT(isValid());
}



TTFSDir::~TTFSDir()
{
    Q_ASSERT(isValid());
}



// -----------------------------------------------------------------------------
// PUBLIC MEMBER FUNCTIONS
// -----------------------------------------------------------------------------

bool TTFSDir::isDir() const
{
    return true;
}



bool TTFSDir::isFile() const
{
    return false;
}



bool TTFSDir::chmod(mode_t mode)
{
    spyINFO("mode:" << mode << toString());
    iError = TTFS::_errUNDEF;
    bool isOk = false;

    isOk = TTFSNode::chmod(mode | LINUX_S_IFDIR);

    if (!isOk)
    {
        spyWARN("TTFSError:" << toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}



bool TTFSDir::rename(const QString& dest)
{
    spyINFO("dest:" << dest << toString());
    iError = TTFS::_errUNDEF;
    bool isOk = false;

    // Check if the directory name ends with '/'.
    if (dest.endsWith('/'))
    {
        isOk = TTFSNode::rename(dest);
    }
    else
    {
        // Bug. Incorrect name for a directory.
        iError = TTFS::errMustBeDir;
        Q_ASSERT(false);
    }

    if (!isOk)
    {
        spyWARN("TTFSError:" << toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}



bool TTFSDir::remove()
{
    spyINFO(toString());
    iError = TTFS::_errUNDEF;
    bool isOk = TTFSNode::remove();
    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}



TTFSFileInfoList TTFSDir::entryList() const
{
    spyINFO(toString());
    Q_ASSERT(isValid());
    iError = TTFS::_errUNDEF;
    errcode_t err = 0;
    bool isOk = false;
    TTFSFileInfoList list;

    if (iFS)
    {
        // Get the (most recent) inode.
        ext2_ino_t inode = inodeNr();
        if (inode)
        {
            typedef std::pair<TTFSFileInfoList*, ext2_filsys> PrivDataType;
            PrivDataType PrivData = std::make_pair(&list, iFS);

            errcode_t err = ext2fs_dir_iterate(iFS, inode, 0, 0, 
                &cbGetFileInfo, (void*) &PrivData);
            if (!err)
            {
                isOk = true;
            }
            else
            {
                // Cannot create a listing of directory specified by iExt2Node.
                list.clear();
                iError = TTFS::errListFailed;
            }
        }
        else
        {
            iError = TTFS::errNodeDoesNotExist;
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN("TTFSError:" << toString() << "errcode_t:" << err);
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    return list;
}



quint64 TTFSDir::size() const
{ 
    quint64 theRetval(0);
    TTFSFileInfoList theList = TTFSDir::entryList();
    foreach (TTFSFileInfo theFileInfo, theList)
    {
        theRetval += theFileInfo.size();
    }
    return theRetval;
}



// -----------------------------------------------------------------------------
// DEBUG MEMBER FUNCTIONS
// -----------------------------------------------------------------------------

#ifndef QT_NO_DEBUG
bool TTFSDir::isValid() const
{
    bool isOk = TTFSNode::isValid();
    Q_ASSERT(isOk = isOk && iNodeType == TTFS::ftDirectory);
    // TODO Q_ASSERT(isOk = isOk && (iPathName.endsWith("/") || iPathName.endsWith("/.") || iPathName.endsWith("/..")));
    return isOk;
}
#endif
