// TTFSDir.h
//
// TTFSDir provides functionality to handle directories within a
// TTFS file system.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#pragma once

// TTFS
#include "TTFSFileInfo.h"
#include "TTFSNode.h"
#include "TTFSTypes.h"

// Qt
#include <QtCore>

class TTFSDir : public TTFSNode
{
public:
    // The ctor of TTFSDir is used to create an object to represent a directory within
    // a TTFS file system.
    //
    // The 'dirPath' parameter specifies the full of the directory (including
    // a trailing '/'). The path should start with the volume identification,
    // which has the format 'extNN:/'.
    //
    // The 'mode' parameter specifies the requested priviliges for operating on the
    // specified directory.
    TTFSDir(
        const QString&  dirPath,                    // Full path of dir, including trailing '/'.
        TTFS::OpenMode  openMode = TTFS::omRead);   // Operating mode.
    ~TTFSDir();

    // Overridden functions from base class.
    bool isDir() const;                     // For TTFSDir: true.
    bool isFile() const;                    // For TTFSDir: false.

    bool chmod(mode_t mode);                // Set chmod for file, mode=ext2-style.
    bool rename(const QString& newName);    // Rename directory.
    bool remove();                          // permission 0644 (rw-r--r--).

    quint64 size() const;
    
    // Listing.
    //
    // entryList()      Provide a directory listing of the current node.
    //                  If the node does not exist, an empty list is returned.
    TTFSFileInfoList entryList() const;        // Provide file/directory listing.

// Private data members.
private:
    // None.

// Unsupported operations.
private:
    TTFSDir();                                      // Unsupported.
    TTFSDir(const TTFSDir&);                        // Unsupported.
    const TTFSDir& operator=(const TTFSDir&);       // Unsupported.

// Debug functions.
public:
#ifndef QT_NO_DEBUG
    bool isValid() const;
#endif
};
