// TTFSError.cpp
//
// Implementation of TTFS::Error functionality.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "TTFSError.h"

// TTFS (this software library)
#include "TTFSDebug.h"

// Qt
#include <QtCore>
#include <QString>

struct TTFSErrorString
{
    TTFS::Error error;  // TTFS::Error code.
    const char* str;    // Readable string.
};

static struct TTFSErrorString sErrorStringList[] =
{
    { TTFS::errCannotBeRoot,            "CannotBeRoot" }, 
    { TTFS::errCannotBeSame,            "CannotBeSame" }, 
    { TTFS::errChmodFailed,             "ChmodFailed" }, 
    { TTFS::errCloseFailed,             "CloseFailed" }, 
    { TTFS::errCreateFailed,            "CreateFailed" }, 
    { TTFS::errCreateiNodeFailed,       "CreateiNodeFailed" }, 
    { TTFS::errDirNotEmpty,             "DirNotEmpty" }, 
    { TTFS::errDirsMustBeSame,          "DirsMustBeSame" }, 
    { TTFS::errFileOpen,                "FileOpen" },
    { TTFS::errFileNotOpen,             "FileNotOpen" }, 
    { TTFS::errFileSystemReadOnly,      "FileSystemReadOnly" }, 
    { TTFS::errGetFileSizeFailed,       "GetFileSizeFailed" }, 
    { TTFS::errFlushFailed,             "FlushFailed" },
    { TTFS::errLinkFailed,              "LinkFailed" }, 
    { TTFS::errListFailed,              "ListFailed" }, 
    { TTFS::errMustBeDir,               "MustBeDir" }, 
    { TTFS::errMustBeFile,              "MustBeFile" }, 
    { TTFS::errNameNotFound,            "NameNotFound" }, 
    { TTFS::errNodeDoesNotExist,        "NodeDoesNotExist" }, 
    { TTFS::errNodeExists,              "NodeExists" },
    { TTFS::errNoTTFS,                  "NoTTFS" },
    { TTFS::errOpenFailed,              "OpenFailed" }, 
    { TTFS::errReadFailed,              "ReadFailed" }, 
    { TTFS::errReadiNodeFailed,         "ReadiNodeFailed" }, 
    { TTFS::errReadBitmapsFailed,       "ReadBitmapsFailed" },
    { TTFS::errSeekFailed,              "SeekFailed" },
    { TTFS::errSetFileSizeFailed,       "SetFileSizeFailed" }, 
    { TTFS::errWritePermRequired,       "WritePermRequired" }, 
    { TTFS::errWrongDestVolume,         "WrongDestVolume" }, 
    { TTFS::errUnknownError,            "UnknownError" }, 
    { TTFS::errUnlinkFailed,            "UnlinkFailed" }, 
    { TTFS::errUnsupportedOperation,    "UnsupportedOperation" },
    { TTFS::errWriteFailed,             "WriteFailed" },
    { TTFS::errWriteiNodeFailed,        "WriteiNodeFailed" }, 
    { TTFS::errWriteBitmapsFailed,      "WriteBitmapsFailed" }, 
    { TTFS::errDirIterateFailed,        "DirIterateFailed" }, 
    { TTFS::errBlockIterateFailed,      "BlockIterateFailed" }, 
    { TTFS::errUnMapBlockFailed,        "UnMapBlockFailed" }, 
    { TTFS::_errUNDEF,                  "<Unknown error code>" }
};



// -----------------------------------------------------------------------------
// DEBUG FUNCTIONS
// -----------------------------------------------------------------------------

#ifndef QT_NO_DEBUG

class TTFSCheckErrorString
{
public:
    TTFSCheckErrorString();     // This class is onyl used to check the validity
    ~TTFSCheckErrorString();    // of the translation table, through the ctor of a 
};                              // static instance of sCheckErrorString of this class.



TTFSCheckErrorString::TTFSCheckErrorString()
{
    // Check the validity of the error string list once.
    static bool done = false;
    if (!done)
    {
        done = true;
        Q_ASSERT(int(TTFS::_errUNDEF) == 0);
        int       i     = 0;
        TTFS::Error error = TTFS::Error(TTFS::_errMIN + 1);
        while (error < TTFS::_errMAX)
        {
            Q_ASSERT(sErrorStringList[i].error == error);
            ++i;
            error = TTFS::Error(int(error) + 1);
        }
    }
}



TTFSCheckErrorString::~TTFSCheckErrorString()
{
    ; // Empty.
}



// Declare a static instance of TTFSCheckErrorString. The ctor will 
// check the array once.
//
static const TTFSCheckErrorString sCheckErrorString;

#endif //QT_NO_DEBUG



// -----------------------------------------------------------------------------
// PUBLIC FUNCTIONS
// -----------------------------------------------------------------------------

QString TTFS::toString(TTFS::Error aError)
{
    Q_ASSERT((TTFS::_errMIN < aError) && (aError < TTFS::_errMAX));

    QString theRetval;
    for (int i = 0; true; ++i)
    {
        if (sErrorStringList[i].error == TTFS::_errUNDEF || 
            sErrorStringList[i].error == aError)
        {
            theRetval = sErrorStringList[i].str;
            break;
        }
    }
    return theRetval;
}



bool TTFS::operator==(bool left, TTFS::Error right) 
{
    return ((left && (right != TTFS::_errUNDEF)) || (!left && (right == TTFS::_errUNDEF)));
}
