// TTFSError.h
//
// TTFSError provides error code functionality for the TTFS library.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#pragma once

// Qt
#include <QtCore>
#include <QString>


// Declare a namespace for all TTFS types, so all identifiers can
// be specified as "TTFS::<name>".
namespace TTFS
{
    // The enum TTFSError enumerates all error codes that are returned from TTFS
    // file system operations.
    enum Error 
    {
        _errUNDEF = 0,                      // For internal use only. Must be 0!
        _errMIN,                            // For internal use only.
        errCannotBeRoot,                    // Specified path should not be root of volume.
        errCannotBeSame,                    // Cannot link file to itself.
        errChmodFailed,                     // Chmod failed.
        errCloseFailed,                     // Closing file failed.
        errCreateFailed,                    // Create file/dir failed.
        errCreateiNodeFailed,               // Cannot create new inode.
        errDirNotEmpty,                     // Dir must be empty before it can be removed.
        errDirsMustBeSame,                  // Source and dest dir for rename must be same.
        errFileOpen,                        // Cannot rename an open file.
        errFileNotOpen,                     // File is not open.
        errFileSystemReadOnly,              // TTFS file system is read-only.
        errGetFileSizeFailed,               // Get file size failed.
        errFlushFailed,                     // Flush file system failed.
        errLinkFailed,                      // Link failed.
        errListFailed,                      // List failed.
        errMustBeDir,                       // Pathname is not a dir.
        errMustBeFile,                      // Pathname is not a regular file.
        errNameNotFound,                    // Cannot find inode for name.
        errNodeDoesNotExist,                // Dir/file does not exist.
        errNodeExists,                      // Dir/file already exists.
        errNoTTFS,                          // Not a TTFS file system.
        errOpenFailed,                      // Open file failed.
        errReadFailed,                      // Read file failed.
        errReadiNodeFailed,                 // Could not read inode information.
        errReadBitmapsFailed,               // Could not read the bitmap information
        errSeekFailed,                      // Seek file failed.
        errSetFileSizeFailed,               // Set file size failed.
        errWritePermRequired,               // Must have write permission for this operation.
        errWrongDestVolume,                 // Cannot hardlink to other volume.
        errUnknownError,                    // Unspecified error occurred.
        errUnlinkFailed,                    // Can't unlink file/dir when removing. 
        errUnsupportedOperation,            // Removing special files in TTFS is not supported.
        errWriteFailed,                     // Write failed.
        errWriteiNodeFailed,                // Could not write inode information.
        errWriteBitmapsFailed,              // Could not write the bitmap information
        errDirIterateFailed,                // failed to iterate directory.
        errBlockIterateFailed,              // failed to iterate blocks.
        errUnMapBlockFailed,                // failed to unmap blocks.
        _errMAX                             // For internal use only.
    };

    // The function toString returns a readable error string for a TTFSError
    // error code. Thia can be used for human readable logging purposes, for example.
    QString toString(Error error);

    // Overload the operator== to make it easy use Error in Q_ASSERT(!error).
    // This means you can define a variable of type TTFS:Error, assign it
    // _errUNDEF, and at some point check Q_ASSERT(!error) to see whether the
    // error has been set to anything else or not. As errors would normally only
    // be set in case of an error, the initialization value would still be kept
    // in the error if no error is encountered.
    //
    // Sample:
    //   error = _errUNDEF;
    //   ... some code that only sets error if an error occurs ...
    //   Q_ASSERT(!error);
    bool operator==(bool left, Error right);
};
