// TTFSFile.cpp
//
// Implementation of TTFSFile class.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "TTFSFile.h"

// TTFS (this software library)
#include "TTFSDebug.h"
#include "TTFSError.h"
#include "TTFSFileSystem.h"
#include "TTFSFileInfo.h"
#include "TTFSIO.h"
#include "TTFSTypes.h"

// Qt
#include <QtCore>
#include <QString>

// ext2fs
#include <ext2fs_prefix.h>
#include <ext2fs/ext2fs.h>

// -----------------------------------------------------------------------------
// CTOR/DTOR
// -----------------------------------------------------------------------------

TTFSFile::TTFSFile(const QString& filePath, TTFS::OpenMode openMode) :
    TTFSNode(filePath, TTFS::ftRegularFile, openMode),
    iNodeNr(0),
    iFile(0),
    iModifiedSize(0),
    iOriginalSize(0)
{
    spyINFO("filePath:" << filePath << "openMode:" << openMode);
    Q_ASSERT(!filePath.endsWith('/'));

    bool isFile(true);
    // Check if the inode is really a file.
    if (iFS)
    {
        ext2_ino_t inode = inodeNr();
        if (inode)
        {
            errcode_t err = ext2fs_check_directory(iFS, inode);
            if (err != EXT2_ET_NO_DIRECTORY)
            {
                spyWARN(filePath << "should be a file, but it is not");
                isFile = false;
            }
        }
    }

    if (!isFile)
    {
        // Clean up. Even though this is programming bug by the caller,
        // we should not continue just issuing an assertion.
        Q_ASSERT(false);
        unmount();
    }
    Q_ASSERT(isValid());
}



TTFSFile::~TTFSFile()
{
    Q_ASSERT(isValid());

    // Close the file. Ok to close if not opened.
    close();
}



// -----------------------------------------------------------------------------
// PUBLIC MEMBER FUNCTIONS
// -----------------------------------------------------------------------------

bool TTFSFile::isDir() const
{
    return false;
}



bool TTFSFile::isFile() const
{
    return true;
}



bool TTFSFile::chmod(mode_t mode)
{
    spyINFO("mode:" << mode << toString());
    bool isOk = false;
    
    isOk = TTFSNode::chmod(mode | LINUX_S_IFREG);

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT(isValid());
    return isOk;
}



bool TTFSFile::rename(const QString& dest)
{
    spyINFO("dest:" << dest << toString());
    iError = TTFS::_errUNDEF;
    bool isOk = false;
    
    // Check if the file is currently open.
    if (iFile)
    {
        if (!dest.endsWith('/'))
        {
            isOk = TTFSNode::rename(dest);
        }
        else
        {
            // Bug. Incorrect name for a file.
            iError = TTFS::errMustBeFile;
            Q_ASSERT(false);
        }
    }
    else
    {
        iError = TTFS::errFileOpen;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    return isOk;
}



bool TTFSFile::remove()
{
    spyINFO("file:" << fullPathName());
    bool isOk = false;
    if (iFile)
    {
        iError = TTFS::errFileOpen;
    }
    else
    {
        isOk = TTFSNode::remove();
    }
    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}



bool TTFSFile::open(TTFS::OpenMode openMode)
{
    spyINFO("openMode:" << openMode << "file:" << fullPathName());
    iError = TTFS::_errUNDEF;
    bool isOk = true;
    iOpenMode = openMode;
    iModifiedSize = iOriginalSize = 0; 

    // Check if file already open.
    if (!iFile)
    {
        // File is not open. Get the (most recent) inode.
        // Might return 0 if the file does not exist.
        ext2_ino_t inode = inodeNr();
        if (inode)
        {
            iNodeNr = inode;
        }
        else
        {
            // The file does not exist yet. 
            if (iOpenMode & TTFS::omWrite)
            {
                isOk = create();
                if (isOk)
                {
                    // Cache the newly created inode.
                    iNodeNr = inodeNr();
                }
                else
                {
                    iError = TTFS::errCreateFailed;
                }
            }
            else
            {
                isOk = false;
                iError = TTFS::errWritePermRequired;
            }
        }
        Q_ASSERT((!isOk && !iNodeNr) || (isOk && iNodeNr));
        // Now, open the file.
        if (isOk)
        {
            isOk = false;
            ext2_inode iNodeStruct;
            errcode_t err = ext2fs_read_inode(iFS, iNodeNr, &iNodeStruct);
            if (!err)
            {
                int flags = 0;
                if (iOpenMode & TTFS::omWrite)
                {
                    flags = (EXT2_FILE_WRITE | EXT2_FILE_CREATE);
                    // Mark TTFS as dirty, flush needed later.
                    markDirty();
                }
                err = ext2fs_file_open2(iFS, iNodeNr, &iNodeStruct, flags, &iFile);
                if (!err)
                {
                    if (iFile)
                    {
                        // All ok, file opened.
                        isOk = true;
                    }
                    else
                    {
                        // This should not happen. No error was reported, bnut iFile invalid.
                        iError = TTFS::errUnknownError;
                        Q_ASSERT(false);
                    }

                }
                else
                {
                    iError = TTFS::errOpenFailed;
                }
            }
            else
            {
                iError = TTFS::errReadiNodeFailed;
            }
        }
    }
    else
    {
        // File already open.
        iError = TTFS::errFileOpen;
        Q_ASSERT(iNodeNr);
    }

    if (isOk)
    {
        iModifiedSize = iOriginalSize = size(); 
    }
    else
    {
        spyWARN(toString());
    }
    Q_ASSERT((!isOk && !iFile) || (isOk && iFile));
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}



bool TTFSFile::close()
{
    spyINFO("file:" << fullPathName());
    iError = TTFS::_errUNDEF;
    bool isOk = false;

    // Check if file is open.
    if (iFile)
    {
        if (iModifiedSize != iOriginalSize)
        {
            setSize(iModifiedSize);
            iOriginalSize = iModifiedSize;
        }
        errcode_t err = ext2fs_file_close(iFile);
        if (!err)
        {
            isOk = true;
        }
        else
        {
            // Close file failed. Should never happen.
            iError = TTFS::errCloseFailed;
            Q_ASSERT(false);
        }
        iFile   = 0;
        iNodeNr = 0;

        TTFSNode::flush();
    }
    else
    {
        isOk = true;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return true;
}



bool TTFSFile::isOpen() const
{
    spyINFO("file:" << fullPathName());
    Q_ASSERT(isValid());
    iError = TTFS::_errUNDEF;

    // Return true if iFile is open.
    return iFile;
}



bool TTFSFile::read(void* buffer, unsigned int wantBytes, unsigned int& gotBytes)
{
    spyINFO("buffer:" << (buffer ? 1 : 0) << "file:" << fullPathName() << "wantBytes:" << wantBytes);
    Q_ASSERT(buffer);
    bool isOk   = false;
    iError       = TTFS::_errUNDEF;
    gotBytes    = 0;

    // Check if the file is open (can only be open if iFS exists).
    if (iFile)
    {
        // Read from file into buffer.
        errcode_t err = ext2fs_file_read(iFile, buffer, wantBytes, &gotBytes);
        if (err)
        {
            // Error while reading file.
            spyWARN("ext2fs_file_read, errcode_t:" << err <<
                "wantBytes:" << wantBytes << "gotBytes:" << gotBytes << "file:" << fullPathName());
            iError = TTFS::errReadFailed;
            isOk = false;
        }
        else
        {
            // No error has been return thus all OK.  
            // WantBytes can always be greater or equal to gotBytes
            isOk = true;
        }
    }
    else
    {
        iError = TTFS::errFileNotOpen;
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}



void TTFSFile::flush()
{
    spyINFO("file:" << fullPathName());
    iError = TTFS::_errUNDEF;
    if (iFile && iFlushNeeded)
    {
        errcode_t err = ext2fs_file_flush(iFile); //!! TODO Why is the file flushed immediately?
        if (err)
        {
            spyWARN("ext2fs_file_flush, errcode_t:" << err << "file:" << fullPathName());
            iError = TTFS::errFlushFailed;
        }
    }
    Q_ASSERT(isValid());
}



bool TTFSFile::write(const void* buffer, unsigned int sizeBytes, unsigned int& writtenBytes)
{
    spyINFO("buffer:" << (buffer ? 1 : 0) << "file:" << fullPathName() << "sizeBytes:" << sizeBytes);
    Q_ASSERT(buffer);
    bool isOk       = false;
    iError          = TTFS::_errUNDEF;
    writtenBytes    = 0;

    // Check if the file is open (can only be open if iFS exists).
    if (iFile)
    {
        // Write from file into buffer.
        errcode_t err = ext2fs_file_write(iFile, buffer, sizeBytes, &writtenBytes);
        if (err)
        {
            spyWARN("ext2fs_file_write, errcode_t:" << err <<
                "sizeBytes:" << sizeBytes << "writtenBytes:" << writtenBytes << "file:" << fullPathName());
            iError = TTFS::errWriteFailed;
        }
        else
        {
            isOk = true;
            iModifiedSize += writtenBytes;
        }
    }
    else
    {
        iError = TTFS::errFileNotOpen;
    }

    Q_ASSERT(isValid());
    return writtenBytes;
}



quint64 TTFSFile::tell() const
{
    spyINFO("file:" << fullPathName());
    Q_ASSERT(isValid());
    iError = TTFS::_errUNDEF;
    quint64 filePos(0);

    // Check if the file is open (can only be open if iFS exists).
    if (iFile)
    {
        ext2_off_t result = 0;
        errcode_t err = ext2fs_file_lseek(iFile, 0, EXT2_SEEK_CUR, &result);
        if (err)
        {
            spyWARN("ext2fs_file_lseek, errcode_t:" << err << "file:" << fullPathName());
            iError = TTFS::errSeekFailed;
        }
        else
        {
            filePos = (quint64)result;
        }
    }
    else
    {
        iError = TTFS::errFileNotOpen;
    }
    return filePos;
}



bool TTFSFile::seek(quint64 pos, int how)
{
    spyINFO("pos:" << pos << "how:" << how << "file:" << fullPathName());
    Q_ASSERT((how == SEEK_SET) || (how == SEEK_CUR) || (how == SEEK_END));
    Q_ASSERT((how != SEEK_END) || (pos <= 0));
    bool isOk = false;
    iError = TTFS::_errUNDEF;

    // Check if the file is open (can only be open if iFS exists).
    if (iFile)
    {
        int whence = EXT2_SEEK_SET;
        switch (how)
        {
            case SEEK_SET:
                whence = EXT2_SEEK_SET; 
                break;

            case SEEK_CUR:
                whence = EXT2_SEEK_CUR;
                break;

            case SEEK_END:
                whence = EXT2_SEEK_END; 
                break;

            default:
                Q_ASSERT(false);
                break;
        }

        ext2_off_t result = 0;
        errcode_t err = ext2fs_file_lseek(iFile, (ext2_off_t)pos, whence, &result);
        if (err)
        {
            spyWARN("ext2fs_file_lseek, errcode_t:" << err <<
                "pos:" << pos << "how:" << how << "file:" << fullPathName());
            iError = TTFS::errSeekFailed;
        }
        else
        {
            Q_ASSERT(result <= size());
            isOk = true;
        }
    }
    else
    {
        iError = TTFS::errFileNotOpen;
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}



quint64 TTFSFile::size() const
{
    spyINFO("file:" << fullPathName());
    Q_ASSERT(isValid());

    return (quint64)ext2fs_file_get_size(iFile);
}



bool TTFSFile::setSize(quint64 aSize)
{
    spyINFO("size:" << aSize << "file:" << fullPathName());
    iError     = TTFS::_errUNDEF;
    bool isOk   = false;
    bool isOpen = false;

    // Check if the file was open.
    if (!iFile)
    {
        // If not, open it now, close it later.
        isOpen = open(TTFS::omRead);
        if (!isOpen)
        {
            iError = TTFS::errNodeDoesNotExist;
        }
    }

    // If the file was succefully opened, get the file size.
    if (iFile)
    {
        errcode_t err = ext2fs_file_set_size(iFile, (ext2_off_t) aSize);
        if (err)
        {
            spyWARN("ext2fs_file_get_size, errcode_t:" << err <<
                "size:" << aSize << "file:" << fullPathName());
            iError = TTFS::errSetFileSizeFailed;
        }
        else
        {
            isOk = true;
        }
    }

    // If the file was closed at the start, close it again.
    if (isOpen)
    {
        isOk = close();
        if (!isOk)
        {
            iError = TTFS::errCloseFailed;
        }
    }
    Q_ASSERT((isOk && !iError) || (!isOk && !iError));
    Q_ASSERT(isValid());
    return isOk;
}



bool TTFSFile::canRead() const
{
    spyINFO("file:" << fullPathName());
    Q_ASSERT(isValid());

    bool allowRead = false;
    if (iFile)
    {
        allowRead = isOpen();
    }
    Q_ASSERT(!allowRead || iFile);
    return allowRead;
}



bool TTFSFile::canWrite() const
{
    spyINFO("file:" << fullPathName());
    Q_ASSERT(isValid());

    bool allowWrite = false;
    if (iFile)
    {
        if (iOpenMode & TTFS::omWrite)
        {
            allowWrite = true;  //!! TODO Should check if TTFS container is writable as well?
        }
    }
    Q_ASSERT(!allowWrite || (iFile && (iOpenMode & TTFS::omWrite)));
    return allowWrite;
}



// -----------------------------------------------------------------------------
// DEBUG MEMBER FUNCTIONS
// -----------------------------------------------------------------------------

#ifndef QT_NO_DEBUG
bool TTFSFile::isValid() const
{
    bool isOk = TTFSNode::isValid();
    Q_ASSERT(isOk = isOk && (iFS || (!iFile && !iNodeNr)));
    Q_ASSERT(isOk = isOk && ((!iFile && !iNodeNr) || (iFile && iNodeNr)));
    Q_ASSERT(isOk = isOk && !iPathName.endsWith("/"));
    return isOk;
}
#endif
