// TTFSFile.h
//
// TTFSFile provides functionality to handle files within a
// TTFS file system.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#pragma once

// TTFS
#include "TTFSNode.h"
#include "TTFSTypes.h"

// Qt
#include <QtCore>



class TTFSFile : public TTFSNode
{
public:
    // The ctor of TTFSFile is used to create an object to represent a file within
    // a TTFS file system.
    //
    // The 'filePath' parameter specifies the full path of the file (not including
    // a trailing '/'). The path should start with the volume identification,
    // which has the format 'extNN:/'.
    //
    // The 'mode' parameter specifies the requested priviliges for operating on the
    // specified file.
    TTFSFile(
        const QString&  filePath,                   // Full path of file, no trailing '/'.
        TTFS::OpenMode  openMode = TTFS::omRead);   // Operating mode.
    ~TTFSFile();

    // Overridden functions from base class.
    virtual bool isDir() const;                     // For TTFSFile: false.
    virtual bool isFile() const;                    // For TTFSFile: true.

    virtual bool chmod(mode_t mode);                // Set chmod for file, mode=ext2-style.
    virtual bool rename(const QString& newName);    // Rename a file.
    virtual bool remove();                          // Remove a file.
    
    // Opening/closing a file.
    bool open(TTFS::OpenMode openMode);     // Open the file.
    bool close();                           // Close the file.
    bool isOpen() const;                    // Is the file open?

    // File size related.
    quint64     size() const;               // Return file size in bytes.
    bool        setSize(quint64 aSize);     // Update file size.

    // File modes.
    bool canRead() const;                   // Is true iff able to read from file.
    bool canWrite() const;                  // Is true iff able to write to file.

    // Operations that require a file to be opened.
    quint64     tell() const;               // Current file position in file.
    bool        seek(                       // Set file position in file. 
        quint64 pos,                        // Absolute/relative file position.
        int how = SEEK_SET);                // Seek mode: SEEK_SET:0, SEEK_CURR:1, SEEK_END:2

    bool read(                              // Read wantBytes from file.
        void* buffer,                       // Buffer owned by caller.
        unsigned int wantBytes,             // Number of bytes to read.
        unsigned int& gotBytes);            // Number of bytes actually read. error if < wantBytes.

    bool write(                             // Write sizeBytes to file. 
        const void* buffer,                 // Buffer owned by caller.
        unsigned int sizeBytes,             // Number of bytes to write.
        unsigned int& writtenBytes);        // Number of bytes actually written. Error if < sizeBytes.

    void flush();
    
// Private data members.
private:
    ext2_ino_t  iNodeNr;                    // iNode number, only cached while file is open.
    ext2_file_t iFile;                      // File pointer, !0 while file is open.
    quint64     iModifiedSize;              // Updates file size used while writing.
    quint64     iOriginalSize;              // Original file size before writing.

// Unsupported operations.
private:
    TTFSFile();                                     // Unsupported.
    TTFSFile(const TTFSFile&);                      // Unsupported.
    const TTFSFile& operator=(const TTFSFile&);     // Unsupported.

// Debug functions.
public:
#ifndef QT_NO_DEBUG
    bool isValid() const;
#endif
};
