// TTFSFileEngine.cpp
//
// Implementation of the TTFSFileEngine class.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "TTFSFileEngine.h"

// TTFS (this software library)
#include "TTFSDir.h"
#include "TTFSFile.h"
#include "TTFSFileSystem.h"
#include "TTFSFileEngineIterator.h"
#include "TTFSFileInfo.h"
#include "TTFSNode.h"
#include "TTFSTypes.h"

#include <QtCore>
#include <QAbstractFileEngine>
#include <QFileInfo>
#include <QString>
#include <QStringList>
#include <QDateTime>

// TTFSTypes.h
const QString TTFS::ContainerFileName("TOMTOM.000");

static const QString kVolumeFormat("ext00:/");
static int kVolumePrefixLength = 3;
static int kVolumeColonPosition = 5;

// -----------------------------------------------------------------------------
// CTOR/DTOR
// -----------------------------------------------------------------------------

TTFSFileEngine::TTFSFileEngine() :
    QAbstractFileEngine(),
    iNode(0)
{
    Q_ASSERT(isValid());
}

TTFSFileEngine::TTFSFileEngine(const QString &aPath) :
    QAbstractFileEngine(),
    iNode(0)
{
    setFileName(aPath);
    Q_ASSERT(isValid());
}

TTFSFileEngine::~TTFSFileEngine()
{
    if (iNode)
    {
        delete iNode;
        iNode = 0;
    }
    Q_ASSERT(isValid());
}

// -----------------------------------------------------------------------------
// PUBLIC MEMBER FUNCTIONS.
// -----------------------------------------------------------------------------

bool TTFSFileEngine::chown(quint16 aUid, quint16 aGid)
{
    Q_ASSERT(isValid());
    bool isOk = false;
    if (iNode)
    {
        isOk = iNode->chown(aUid, aGid);
    }
    return isOk;
}

bool TTFSFileEngine::hardLink(const QString& newPath)
{
    Q_ASSERT(isValid());
    bool isOk = false;
    if (iNode)
    {
        QString thePath(iNode->fullPathName());
        TTFSFileInfo theNewPathInfo(newPath);
        if (iNode->exists() && !theNewPathInfo.Exists())
        {
            if (iNode->isFile())
            {
                isOk = TTFSFile(thePath, TTFS::omWrite).hardLink(newPath);
            }
            else if (iNode->isDir())
            {
                isOk = TTFSDir(thePath, TTFS::omWrite).hardLink(newPath);
            }
        }
    }
    return isOk;
}

quint64 TTFSFileEngine::freeBytes() const
{
    Q_ASSERT(isValid());
    quint64 theRetval(0);
    if (iNode && iNode->isDir())
    {
        theRetval = iNode->freeBytes();
    }
    return theRetval;
}

bool TTFSFileEngine::chmod(quint16 mode)
{
    bool isOk(false);
    if (iNode)
    {
        isOk = iNode->chmod((mode_t)mode);
    }
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSFileEngine::setTime(time_t timeMod, time_t timeAccess)
{
    bool isOk(false);
    if (iNode)
    {
        isOk = iNode->setTime(timeMod, timeAccess);
    }
    Q_ASSERT(isValid());
    return isOk;
}

quint64 TTFSFileEngine::usedBytes() const
{
    Q_ASSERT(isValid());
    quint64 theRetval(0);
    if (iNode && iNode->isDir())
    {
        theRetval = iNode->usedBytes();
    }
    return theRetval;
}

// -----------------------------------------------------------------------------
// PUBLIC STATIC MEMBER FUNCTIONS.
// -----------------------------------------------------------------------------

QFileInfoList TTFSFileEngine::volumes()
{
    QFileInfoList theFiles;
    TTFSFileSystem& theInstance = TTFSFileSystem::GetInstance();
    QStringList theVolumes(theInstance.GetVolumes());
    foreach (QString theVolume, theVolumes)
    {
        theFiles.push_back(QFileInfo(theVolume.append("/")));
    }
    return theFiles;
}

// -----------------------------------------------------------------------------
// INHERITED/OVERRIDDEN PUBLIC MEMBER FUNCTIONS.
// -----------------------------------------------------------------------------

bool TTFSFileEngine::atEnd() const
{
    bool theRetval(false);
    Q_ASSERT(isValid());
    if (iNode && iNode->isFile())
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        theRetval = (quint64)theFile->tell() == (quint64)theFile->size(); // TODO Cache
    }
    return theRetval;
}

QAbstractFileEngine::Iterator* TTFSFileEngine::beginEntryList(
                                                QDir::Filters aFilters,
                                                const QStringList &aFilterNames)
{
    return new TTFSFileEngineIterator(aFilters, aFilterNames);
}

QAbstractFileEngine::Iterator* TTFSFileEngine::endEntryList()
{
    return 0;
}

bool TTFSFileEngine::close()
{
    bool theRetval(true);
    if (iNode && iNode->isFile())
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        theRetval = theFile->close();
    }
    Q_ASSERT(isValid());
    return theRetval;
}

bool TTFSFileEngine::caseSensitive() const
{
    Q_ASSERT(isValid());
    return true;
}

bool TTFSFileEngine::copy(const QString& aNewName)
{
    bool isOk = QAbstractFileEngine::copy(aNewName);
    Q_ASSERT(isValid());
    return isOk;
}

QStringList TTFSFileEngine::entryList(QDir::Filters filters,
    const QStringList& filterNames) const
{
    Q_ASSERT(isValid());
    return QAbstractFileEngine::entryList(filters, filterNames);
}

bool TTFSFileEngine::extension(Extension aExtension, 
                                 const ExtensionOption* aOption,
                                 ExtensionReturn* aOutput)
{
    Q_UNUSED(aOption);
    Q_UNUSED(aOutput);
    bool theRetval(false);
    Q_ASSERT(isValid());
    if (QAbstractFileEngine::AtEndExtension == aExtension)
    {
        theRetval = atEnd();
    }
    return theRetval;
}

QAbstractFileEngine::FileFlags TTFSFileEngine::fileFlags(FileFlags aType) const
{
    Q_ASSERT(isValid());
    QAbstractFileEngine::FileFlags theRetval = 0;
    if (iNode)
    {
        theRetval = TTFSFileEngine::GetFileFlags(iNode->fullPathName(), aType);
    }
    return theRetval;
}

QAbstractFileEngine::FileFlags TTFSFileEngine::GetFileFlags(const QString& aPath, FileFlags aType)
{
    QAbstractFileEngine::FileFlags flags = 0;
    // test if there is an EXT3 image mounted with the provided volume id.
    TTFSFileSystem& theInstance = TTFSFileSystem::GetInstance();
    ext2_filsys theFS = theInstance.GetRootEXT2(aPath);
    if (theFS)
    {
        // test if it is directory.
        QString thePath(aPath);
        if (thePath.endsWith(":"))
        {
            thePath.append("/");
        }
        TTFSFileInfo theFileInfo(thePath);
        if (theFileInfo.isDir())
        {
            if (aType & QAbstractFileEngine::TypesMask)
            {
                flags |= QAbstractFileEngine::DirectoryType;
            }
            if (aType & QAbstractFileEngine::FlagsMask)
            {
                flags |= ExistsFlag;
                QString theRoot(TTFSFileSystem::GetVolume(aPath));
                if (!theRoot.endsWith("/"))
                {
                    theRoot.append("/");
                }
                if (QString::compare(aPath, theRoot) == 0)
                {
                    flags |= QAbstractFileEngine::RootFlag;
                }
            }
            if (aType & QAbstractFileEngine::PermsMask)
            {
                bool readOnly(theInstance.IsReadOnly(aPath));
                flags |= convertToFileFlags(theFileInfo, readOnly);
            }
        }
        else
        {
            theFileInfo = TTFSFileInfo(aPath);
            if (theFileInfo.isFile())
            {
                if (aType & QAbstractFileEngine::TypesMask)
                {
                    flags |= QAbstractFileEngine::FileType;
                }
                if (aType & QAbstractFileEngine::FlagsMask)
                {
                    flags |= ExistsFlag;
                }
                if (aType & QAbstractFileEngine::PermsMask)
                {
                    // on ext3 file system user and owner has no difference.
                    bool readOnly(theInstance.IsReadOnly(aPath));
                    flags |= convertToFileFlags(theFileInfo, readOnly);
                }
            }
        }
    }
    return flags;
}

QString TTFSFileEngine::fileName(FileName file) const
{
    Q_ASSERT(isValid());
    QString theRetval;
    if (iNode)
    {
        theRetval = iNode->fullPathName();
    }
    switch (file)
    {
    case BaseName:          // The name of the file excluding the path.
        theRetval = theRetval.mid(theRetval.lastIndexOf("/") + 1);
        break;
    case PathName:          // The path to the file excluding the base name.
        theRetval = theRetval.mid(0, theRetval.lastIndexOf("/"));
        break;
    case AbsoluteName:      // The absolute path to the file (including the base name).
        // nothing to do.
        break;
    case AbsolutePathName:  // The absolute path to the file (excluding the base name).
        theRetval = theRetval.mid(0, theRetval.lastIndexOf("/"));
        break;
    case LinkName:          // The full file name of the file that this file is a link to.
                            // (This will be empty if this file is not a link.)
        theRetval = QString();
        break;
    case CanonicalName:     // Often very similar to LinkName.
                            // Will return the true path to the file.
                            // (This will be empty if this file is not a link.)
        theRetval = QString();
        break;
    case CanonicalPathName: // Same as CanonicalName, excluding the base name.
        theRetval = QString();
        break;
    case BundleName:        // Returns the name of the bundle implies BundleType is set.
        theRetval = QString();
        break;
    default:
        break;
    }
    return theRetval;
}

QDateTime TTFSFileEngine::fileTime(FileTime time) const
{
    Q_ASSERT(isValid());
    QDateTime result;
    if (iNode && (iNode->isDir() || iNode->isFile()))  // Does not work on links, pipes, sockets...
    {
        TTFSFileInfo fileInfo(iNode->fullPathName());
        switch (time)
        {
        case CreationTime:
            {
                unsigned int timeChanged = fileInfo.GetChangeTime();
                if (timeChanged == 0)
                {
                    timeChanged = fileInfo.GetModifiedTime();
                }
                result.setTime_t(timeChanged);
            }
            break;

        case ModificationTime:
            result.setTime_t(fileInfo.GetModifiedTime());
            break;

        case AccessTime:
            result.setTime_t(fileInfo.GetAccessTime());
            break;

        default:
            Q_ASSERT(false);
            break;
        }
    }
    return result;
}

bool TTFSFileEngine::flush()
{
    if (iNode && iNode->isFile())
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        theFile->flush();
    }
    Q_ASSERT(isValid());
    return true;
}

int TTFSFileEngine::handle() const
{
    Q_ASSERT(isValid());
    return QAbstractFileEngine::handle();
}

/*static*/
bool TTFSFileEngine::canHandle(const QString& pathName)
{
    return (pathName.length() > kVolumeColonPosition) &&
        (pathName.at(kVolumeColonPosition) == kVolumeFormat.at(kVolumeColonPosition)) &&
        (pathName.left(kVolumePrefixLength) == kVolumeFormat.left(kVolumePrefixLength));
}

bool TTFSFileEngine::isRelativePath() const
{
    Q_ASSERT(isValid());
    bool theRetval(false);
    if (iNode)
    {
        theRetval = !TTFSFileEngine::canHandle(iNode->fullPathName());
    }
    return theRetval;
}

bool TTFSFileEngine::isSequential() const
{
    Q_ASSERT(isValid());
    return false;
}

bool TTFSFileEngine::link(const QString& newName)
{
    bool isOk = QAbstractFileEngine::link(newName);
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSFileEngine::mkdir(const QString& dirName, bool createParentDirectories) const
{
    Q_ASSERT(isValid());
    Q_UNUSED(createParentDirectories);  //!! TODO: should be used!

    bool isOk(false);
    QString theDirName = QDir::fromNativeSeparators(dirName);
    TTFSDir theDirectory(theDirName, TTFS::omWrite);
    if (!theDirectory.exists())
    {
        isOk = theDirectory.create();
    }
    return isOk;
}

bool TTFSFileEngine::open(QIODevice::OpenMode mode)
{
    bool isOk(true);
    QString thePath;
    bool theIsFile(false);
    Q_ASSERT(iNode);

    theIsFile = iNode->isFile();
    thePath = iNode->fullPathName();
    if (theIsFile)
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        if (theFile->isOpen())
        {
            theFile->close();
        }
    }

    if (mode & QIODevice::Truncate && iNode->exists())
    {
        TTFSFile theFile(thePath, TTFS::omWrite);
        theFile.remove();
    }
    TTFS::OpenMode theOpenMode(TTFS::omRead);
    if ((mode & QIODevice::WriteOnly) &&
        (mode & QIODevice::ReadOnly))
    {
        theOpenMode = TTFS::omReadWrite;
    }
    else if (mode & QIODevice::WriteOnly)
    {
        theOpenMode = TTFS::omWrite;
    }
    if (mode & QIODevice::Append)
    {
        theOpenMode = TTFS::omAppend;
    }
    else if (mode & QIODevice::Text)
    {
        Q_ASSERT(false && "Not implemented yet!");
    }
    if (theOpenMode == TTFS::omRead)
    {
        // Check whether the file exists or not.
        // If the file doesn't exist no need to open file.
        TTFSFileInfo theFileInfo(thePath);
        isOk = theFileInfo.Exists() && theFileInfo.isFile();
    }
    if (isOk)
    {
        TTFSFile* theFile = new TTFSFile(thePath);
        isOk = theFile->open(theOpenMode);
        if (isOk)
        {
            delete iNode;
            iNode = theFile;
        }
        else
        {
            delete theFile;
        }
    }
    Q_ASSERT(isValid());
    return isOk;
}

QString TTFSFileEngine::owner(FileOwner owner) const
{
    Q_ASSERT(isValid());
    return QAbstractFileEngine::owner(owner);
}

uint TTFSFileEngine::ownerId(FileOwner owner) const
{
    Q_ASSERT(isValid());
    return QAbstractFileEngine::ownerId(owner);
}

qint64 TTFSFileEngine::pos() const
{
    Q_ASSERT(isValid());
    qint64 result = 0;
    if (iNode && iNode->isFile())
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        result = (qint64)theFile->tell();
    }
    return result;
}

qint64 TTFSFileEngine::read(char* data, qint64 maxlen)
{
    qint64 theRetval = 0;
    if (iNode && iNode->isFile())
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        unsigned int theBytesRead;
        if (theFile->read(data, (unsigned int)maxlen, theBytesRead))
        {
            theRetval = (qint64)theBytesRead;
        }
    }
    Q_ASSERT(isValid());
    return theRetval;
}

qint64 TTFSFileEngine::readLine(char* data, qint64 maxlen)
{
    // NOT IMPLEMENTED.
    Q_ASSERT(false);
    bool isOk = QAbstractFileEngine::readLine(data, maxlen);
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSFileEngine::remove()
{
    bool isOk(false);
    if (iNode)
    {
        QString thePath(iNode->fullPathName());
        if (iNode->isFile())
        {
            TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
            if (theFile->isOpen())
            {
                theFile->close();
            }

            TTFSFile theFileToRemove(thePath, TTFS::omWrite);
            isOk = theFileToRemove.remove();
        }
        else if (iNode->isDir())
        {
            TTFSDir theDirToRemove(thePath, TTFS::omWrite);
            isOk = theDirToRemove.remove();
        }
    }
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSFileEngine::rename(const QString& newName)
{
    bool isOk(false);
    if (iNode && iNode->exists())
    {
        if (iNode->isFile())
        {
            isOk = TTFSFile(iNode->fullPathName(), TTFS::omWrite).rename(newName);
        }
        else if (iNode->isDir())
        {
            isOk = TTFSDir(iNode->fullPathName(), TTFS::omWrite).rename(newName);
        }
    }
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSFileEngine::rmdir(const QString& dirName, bool recurseParentDirectories) const
{
    Q_ASSERT(isValid());
    bool isOk = false;
    QString dir = QDir::fromNativeSeparators(dirName);
    if (dir.endsWith("/"))
    {
        dir.chop(1);
    }
    bool isRemoved = false;
    do
    {
        TTFSDir theDirectory(dir, TTFS::omWrite);
        isRemoved = theDirectory.remove();
        isOk |= isRemoved;
        if (!recurseParentDirectories)
        {
            break;
        }
        int index = dir.lastIndexOf("/");
        if (index < 0)
        {
            break;
        }
        dir = dir.mid(0, index);
    }
    while (isRemoved);
    return isOk;
}

bool TTFSFileEngine::seek(qint64 offset)
{
    bool isOk(false);
    if (iNode && iNode->isFile())
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        isOk = theFile->seek(offset);
    }
    Q_ASSERT(isValid());
    return isOk;
}

void TTFSFileEngine::setFileName(const QString& aPath)
{
    QString thePath(QDir::fromNativeSeparators(aPath));
    if (thePath.endsWith(":"))
    {
        thePath += "/";
    }
    QAbstractFileEngine::FileFlags theFlags =
        TTFSFileEngine::GetFileFlags(thePath, QAbstractFileEngine::TypesMask);
    // Directories that are being created will not have the Directory type
    // Thus if the path ends in a / it is also a directory
    if ((theFlags & QAbstractFileEngine::DirectoryType) != 0 || thePath.endsWith("/"))
    {
        iNode = new TTFSDir(thePath);
    }
    else
    {
        // Assume it is a file. If it does not exist yet, so be it.
        // Thst file can be created using |open()|.
        // A folder on the other hand can be created using |mkdir()|.
        iNode = new TTFSFile(aPath);
    }
    Q_ASSERT(isValid());
}

bool TTFSFileEngine::setPermissions(uint perms)
{
    // Need to transform Qt permissions -> Linux/ext2 fs mode permissions (0x640 -> 0640).
    // Qt permissions are readable as hex values;
    // Linux/ext2 fs mode permissions are readable as octal;
    bool isOk(true);
    
    if (perms) 
    {
        quint16 mode = 0;
        
        QFile::Permissions thePermissions;
        if ((perms & QAbstractFileEngine::ReadOwnerPerm) || (perms & QAbstractFileEngine::ReadUserPerm))
        {
            mode |= LINUX_S_IRUSR;
        }
        if ((perms & QAbstractFileEngine::WriteOwnerPerm) || (perms & QAbstractFileEngine::WriteUserPerm))
        {
            mode |= LINUX_S_IWUSR;
        }
        if ((perms & QAbstractFileEngine::ExeOwnerPerm) || (perms & QAbstractFileEngine::ExeUserPerm))
        {
            mode |= LINUX_S_IXUSR;
        }
        if (perms & QAbstractFileEngine::ReadGroupPerm)
        {
            mode |= LINUX_S_IRGRP;
        }
        if (perms & QAbstractFileEngine::WriteGroupPerm)
        {
            mode |= LINUX_S_IWGRP;
        }
        if (perms & QAbstractFileEngine::ExeGroupPerm)
        {
            mode |= LINUX_S_IXGRP;
        }
        if (perms & QAbstractFileEngine::ReadOtherPerm)
        {
            mode |= LINUX_S_IROTH;
        }
        if (perms & QAbstractFileEngine::WriteOtherPerm)
        {
            mode |= LINUX_S_IWOTH;
        }
        if (perms & QAbstractFileEngine::ExeOtherPerm)
        {
            mode |= LINUX_S_IXOTH;
        }
        isOk = chmod(mode);
    }
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSFileEngine::setSize(qint64 size)
{
    bool isOk(false);
    if (iNode && iNode->isFile())
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        isOk = theFile->setSize(size);
    }
    Q_ASSERT(isValid());
    return isOk;
}

qint64 TTFSFileEngine::size() const
{
    qint64 size = 0;
    if (iNode && iNode->isFile())
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        if (!theFile->isOpen())
        {
            theFile = new TTFSFile(theFile->fullPathName());
            if (theFile->open(TTFS::omRead))
            {
                size = (qint64)theFile->size();
                theFile->close();
            }
            delete theFile;
        }
        else
        {
            size = (qint64)theFile->size();
        }
    }
    Q_ASSERT(isValid());
    return size;
}

qint64 TTFSFileEngine::write(const char* aData, qint64 aLen)
{
    qint64 theRetval(0);
    if (iNode && iNode->isFile())
    {
        TTFSFile* theFile = static_cast<TTFSFile*>(iNode);
        unsigned int theBytesWritten;
        if (theFile->write(aData, (unsigned long) aLen, theBytesWritten))
        {
            theRetval = (qint64)theBytesWritten;
        }
    }
    Q_ASSERT(isValid());
    return theRetval;
}

bool TTFSFileEngine::supportsExtension(Extension aExtension) const
{
    return (QAbstractFileEngine::AtEndExtension == aExtension);
}

// -----------------------------------------------------------------------------
// PRIVATE HELPER FUNCTIONS.
// -----------------------------------------------------------------------------

QAbstractFileEngine::FileFlags TTFSFileEngine::convertToFileFlags(
    const TTFSFileInfo& fileInfo, bool readOnly)
{
    QAbstractFileEngine::FileFlags flags = 0;

    const int kReadPermission   = 4;
    const int kWritePermission  = 2;
    const int kExePermission    = 1;

    // On an ext2 file system user and owner are the same.
    //!! TODO @@change readOnly - does not affect file permission - 

    unsigned int userPerm = fileInfo.GetUserPermissions();
    if (userPerm & kReadPermission)
    {
        flags |= QAbstractFileEngine::FileFlags(ReadOwnerPerm | ReadUserPerm);
    }
    if (!readOnly && (userPerm & kWritePermission))
    {
        flags |= QAbstractFileEngine::FileFlags(WriteOwnerPerm | WriteUserPerm);
    }
    if (!readOnly && (userPerm & kExePermission))   //!! TODO: Why is readonly not allowed to be exec?
    {
        flags |= QAbstractFileEngine::FileFlags(ExeOwnerPerm | ExeUserPerm);
    }

    unsigned int groupPerm = fileInfo.GetGroupPermissions();
    if (groupPerm & kReadPermission)
    {
        flags |= QAbstractFileEngine::FileFlags(ReadGroupPerm);
    }
    if (!readOnly && (groupPerm & kWritePermission))
    {
        flags |= QAbstractFileEngine::FileFlags(WriteGroupPerm);
    }
    if (!readOnly && (groupPerm & kExePermission))
    {
        flags |= QAbstractFileEngine::FileFlags(ExeGroupPerm);
    }

    unsigned int otherPerm = fileInfo.GetOtherPermissions();
    if (otherPerm & kReadPermission)
    {
        flags |= QAbstractFileEngine::FileFlags(ReadOtherPerm);
    }
    if (!readOnly && (otherPerm & kWritePermission))
    {
        flags |= QAbstractFileEngine::FileFlags(WriteOtherPerm);
    }
    if (!readOnly && (otherPerm & kExePermission))
    {
        flags |= QAbstractFileEngine::FileFlags(ExeOtherPerm);
    }
    return flags;
}

// -----------------------------------------------------------------------------
// DEBUG MEMBER FUNCTIONS.
// -----------------------------------------------------------------------------

#ifndef QT_NO_DEBUG
bool TTFSFileEngine::isValid() const
{
    bool isOk = true;
    Q_ASSERT(isOk = isOk && (!iNode || iNode->isValid()));
    return isOk;
}
#endif
