// TTFSFileEngine.h
//
// The TTFSFileEngine class implements a QAbstractFileEngine. For every
// file object that can be handled by the TTFSFileEngine, an instance
// is created (by the TTFSFileEngineHandler).
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#pragma once

// TTFS
#include "TTFSTypes.h"

// Qt
#include <QAbstractFileEngine>
#include <QFileInfo>
#include <QString>
#include <QStringList>
#include <QDateTime>

class TTFSFileInfo;
class TTFSNode;

class TTFSFileEngine : public QAbstractFileEngine
{
public:
    TTFSFileEngine();
    explicit TTFSFileEngine(const QString& pathName);
    virtual ~TTFSFileEngine();

    // TTFS-specific functions. For a complete explanation of these function,
    // refer to TTFSNode.h.

    bool    chmod(quint16 mode);                // Change attributes.
    bool    chown(quint16 uid, quint16 gid);    // Change owner/group.
    bool    hardLink(const QString& dest);      // Create a hard link.
    bool    setTime(time_t mod, time_t acc);    // Set time modification/last-access time.

    quint64 usedBytes() const;                  // Used bytes in this volume.
    quint64 freeBytes() const;                  // Free bytes in this volume.

    static QFileInfoList volumes();                 // List of all volumes, with trailing '/'.
    static bool canHandle(const QString& pathName); // True if a TTFS engine can handle it.


    // Inherited/overridden functions from QAbstractFileEngine.
    // These functions are called by QFile and QDir as well.

    virtual bool        atEnd() const;
    virtual QAbstractFileEngine::Iterator*   beginEntryList(QDir::Filters filters, const QStringList &filterNames);
    virtual QAbstractFileEngine::Iterator*   endEntryList();
    virtual bool        caseSensitive() const;
    virtual bool        close();
    virtual bool        copy(const QString& newName);
    virtual QStringList entryList(QDir::Filters filters, const QStringList& filterNames) const;
    virtual bool        extension(Extension extension, const ExtensionOption* option = 0, ExtensionReturn* output = 0);
    virtual FileFlags   fileFlags(FileFlags type = FileInfoAll) const;
    virtual QString     fileName(FileName file = DefaultName) const;
    virtual QDateTime   fileTime(FileTime time) const;
    virtual bool        flush();
    virtual int         handle() const;
    virtual bool        isRelativePath() const;
    virtual bool        isSequential() const;
    virtual bool        link(const QString& newName);
    virtual bool        mkdir(const QString& dirName, bool createParentDirectories) const;
    virtual bool        open(QIODevice::OpenMode mode);
    virtual QString     owner(FileOwner owner) const;
    virtual uint        ownerId(FileOwner owner) const;
    virtual qint64      pos() const;
    virtual qint64      read(char* data, qint64 maxlen);
    virtual qint64      readLine(char* data, qint64 maxlen);
    virtual bool        remove();
    virtual bool        rename(const QString& newName);
    virtual bool        rmdir(const QString& dirName, bool recurseParentDirectories) const;
    virtual bool        seek(qint64 offset);
    virtual void        setFileName(const QString& aPath);
    virtual bool        setPermissions(uint perms);
    virtual bool        setSize(qint64 size);
    virtual qint64      size() const;
    virtual bool        supportsExtension(Extension extension) const;
    virtual qint64      write(const char* data, qint64 len);

// Private helper functions.
private:
    static QAbstractFileEngine::FileFlags convertToFileFlags(   // Convert ext2fs persmissions to
        const TTFSFileInfo& fileInfo,                           // Qt persmission.
        bool readOnly);
    static QAbstractFileEngine::FileFlags GetFileFlags(const QString& aPath, FileFlags aType);

// Privata data members.
private:
    TTFSNode*   iNode;          // The current file object.
                                // Note. Caching TTFSNode object to keep the handle
                                // for continous writing/reading operations.
                                // Life cycle of that object is from open() -> close().
                                // Otherwise, no use of that cache.

// Unsupported operations.
private:
    TTFSFileEngine(const TTFSFileEngine&);                  // Unsupported.
    const TTFSFileEngine& operator=(const TTFSFileEngine&); // Unsupported.

// Debug functions.
public:
#ifndef QT_NO_DEBUG
    bool isValid() const;
#endif
};
