// TTFSFileInfo.cpp
//
// Implementation of TTFSFileInfo class type.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "TTFSFileInfo.h"

// TTFS (this software library)
#include "TTFSLookupEntry.h"
#include "TTFSFile.h"
#include "TTFSFileSystem.h"

// Qt
#include <QtCore>

// ext2fs
#if defined(Q_WS_WIN)
#include "ext2WinPrefix.h"
#endif

#if defined(Q_WS_MAC)
#include "ext2MacPrefix.h"
#endif

#include <ext2fs/ext2fs.h>

static const char* kPermissionTable[8] = {
        "---",
        "--x",
        "-w-",
        "-wx",
        "r--",
        "r-x",
        "rw-",
        "rwx",
    };

// Callbacks
extern "C"
{
    /**
      * @brief
      * Looks for specific direntry. If the direntry is found it returns DIRENT_ABORT and 
      * therefore the lookup loop ends.
      */
    int lookup_entry(ext2_dir_entry *dirent, int offset, int blocksize, char *buf, void *private_data);
}

TTFSFileInfo::TTFSFileInfo() :
   iFileType(TTFS::ftUnknown),
   iNodeMode(0)
{
    ; // empty
}

TTFSFileInfo::TTFSFileInfo(const TTFSFileInfo& aFileInfo) :
    iAbsolutePath(aFileInfo.iAbsolutePath),
    iFilename(aFileInfo.iFilename),
    iFileType(aFileInfo.iFileType),
    iFileSize(aFileInfo.iFileSize),
    iNodePtr(aFileInfo.iNodePtr),
    iNodeMode(aFileInfo.iNodeMode),
    iINode(aFileInfo.iINode)
{
    ; // empty
}

TTFSFileInfo::TTFSFileInfo(const QString& aAbsolutePath, 
                                     const QString& aFilename,
                                     TTFS::FileType aFileType,
                                     ext2_ino_t aNodePtr,
                                     ext2_inode aInode) :
    iAbsolutePath(aAbsolutePath),
    iFilename(aFilename),
    iFileType(aFileType),
    iFileSize(aInode.i_size),
    iNodePtr(aNodePtr),
    iNodeMode(aInode.i_mode),
    iINode(aInode)
{
    ; // empty
}

TTFSFileInfo::TTFSFileInfo(const QString& aFilePath,
                                     TTFS::FileType aFileType,
                                     ext2_ino_t aNodePtr,
                                     ext2_inode aInode) :
    iFileType(aFileType),
    iFileSize(aInode.i_size),
    iNodePtr(aNodePtr),
    iNodeMode(aInode.i_mode),
    iINode(aInode)
{
    int pos = aFilePath.lastIndexOf(QLatin1Char('/'));
    if (pos > -1)
    {
        iFilename = aFilePath.mid(pos + 1);
        iAbsolutePath = aFilePath.left(pos);
    }
    else
    {
        iAbsolutePath = QString(QLatin1Char('/'));
        iFilename = aFilePath;
    }
}

TTFSFileInfo::TTFSFileInfo(const QString& aPath):
    iFileType(TTFS::ftUnknown),
    iFileSize(0),
    iNodePtr(0),
    iNodeMode(0)
{
    // Did we find a EXT2 FS??
    ext2_filsys theExt2FS = TTFSFileSystem::GetInstance().GetRootEXT2(aPath);
    if (theExt2FS)
    {
        QString theParentFolder = TTFSNode::parentFolder(aPath);
        QString theEntryName = TTFSNode::nodeName(aPath);
        ext2_ino_t theExt2iNodePtr(0);
        errcode_t err = ext2fs_namei_follow(theExt2FS, 
                                            EXT2_ROOT_INO, EXT2_ROOT_INO,
                                            theParentFolder.toUtf8().data(),
                                            &theExt2iNodePtr);
        // err will be non 0 if the file doesn't exist and
        if (err == 0)
        {
            if (theEntryName.isEmpty())
            {
                ext2_inode theInode;
                err = ext2fs_read_inode(theExt2FS, EXT2_ROOT_INO, &theInode);
                if (err == 0)
                {
                    iFileSize = theInode.i_size;
                    iNodePtr = EXT2_ROOT_INO;
                    iNodeMode = theInode.i_mode;
                    iINode = theInode;
                    iFilename = theEntryName;
                    iAbsolutePath = theParentFolder;
                    iFileType = TTFS::ftDirectory;
                }
            }
            else
            {
                // Looking up for entry
                TTFSLookupEntry theLookupEntry;
                theLookupEntry.iFileSystem = theExt2FS;
                theLookupEntry.iFound = false;
                theLookupEntry.iEntryNameUtf8 = theEntryName.toUtf8();
                err = ext2fs_dir_iterate(theExt2FS, theExt2iNodePtr, 0, 0, 
                    &lookup_entry, (void*)&theLookupEntry);
                if (err == 0 && theLookupEntry.iFound)
                {
                    iNodeMode = theLookupEntry.iNode.i_mode;
                    iFileSize = theLookupEntry.iNode.i_size;
                    iNodePtr = theLookupEntry.iNodeNumber;

                    iINode = theLookupEntry.iNode;
                    iFilename = theEntryName;
                    iAbsolutePath = theParentFolder;

                    if (theLookupEntry.iNode.i_mode & LINUX_S_IFDIR)
                    {
                        iFileType = TTFS::ftDirectory;
                    }
                    else if (theLookupEntry.iNode.i_mode & LINUX_S_IFREG)
                    {
                        iFileType = TTFS::ftRegularFile;
                    }
                    else
                    {
                        ; // TODO: other file types.
                        //LINUX_S_IFSOCK 0140000
                        //LINUX_S_IFLNK     0120000
                        //LINUX_S_IFBLK  0060000
                        //LINUX_S_IFCHR  0020000
                        //LINUX_S_IFIFO  0010000
                        iNodeMode = 0;
                        iNodePtr = 0;
                    }
                }
            }
        }
    }
}

TTFSFileInfo::~TTFSFileInfo()
{
    ; // empty
}

bool TTFSFileInfo::Exists() const
{
    return (iNodeMode != 0);
}

const QString& TTFSFileInfo::GetAbsolutePath() const
{
    return iAbsolutePath;
}

const QString& TTFSFileInfo::GetFilename() const
{
    return iFilename;
}

QString TTFSFileInfo::GetPath() const
{
    return QString(iAbsolutePath).append('/').append(iFilename);
}

bool TTFSFileInfo::isFile() const
{
    return iFileType == TTFS::ftRegularFile;
}

bool TTFSFileInfo::isDir() const
{
    return iFileType == TTFS::ftDirectory;
}

long long TTFSFileInfo::size() const
{
    return iFileSize;
}

void TTFSFileInfo::SetParentDirectory(const QString& aParentDirectory)
{
    iAbsolutePath = aParentDirectory;
}

ext2_ino_t TTFSFileInfo::GetNodeNumber() const
{
    return iNodePtr;
}

#define LINUX_S_IRWXU_OCTET 2
#define LINUX_S_IRWXG_OCTET 1
#define LINUX_S_IRWXO_OCTET 0

QString TTFSFileInfo::GetModeString() const
{
    // It doesn't express the 
    QString theMode;
    unsigned int theFile(iNodeMode & LINUX_S_IFMT);
    switch (theFile)
    {
    case LINUX_S_IFREG:
        theMode.append('-');
        break;
    case LINUX_S_IFDIR:
        theMode.append('d');
        break;
    case LINUX_S_IFBLK:
        theMode.append('b');
        break;
    case LINUX_S_IFCHR:
        theMode.append('c');
        break;
    case LINUX_S_IFLNK:
        theMode.append('l');
        break;
    case LINUX_S_IFIFO:
        theMode.append('p');
        break;
    case LINUX_S_IFSOCK:
        theMode.append('s');
        break;
    default:
        theMode.append(' ');
        break;
    }
    unsigned int theUser(iNodeMode & LINUX_S_IRWXU);
    theUser >>= 3 * LINUX_S_IRWXU_OCTET;
    if (theUser >= 8)
    {
        theUser = 0;
    }
    theMode.append(kPermissionTable[theUser]);

    unsigned int theGroup(iNodeMode & LINUX_S_IRWXG);
    theGroup >>= 3 * LINUX_S_IRWXG_OCTET;
    if (theGroup >= 8)
    {
        theGroup = 0;
    }
    theMode.append(kPermissionTable[theGroup]);

    unsigned int theOthers(iNodeMode & LINUX_S_IRWXO);
    theOthers >>= 3 * LINUX_S_IRWXO_OCTET;
    if (theOthers >= 8)
    {
        theOthers = 0;
    }
    theMode.append(kPermissionTable[theOthers]);

    return theMode;
}

unsigned int TTFSFileInfo::GetOwnerPermissions() const
{
    // Under ext3 file system it is the same as user permissions.
    return GetUserPermissions();
}

unsigned int TTFSFileInfo::GetUserPermissions() const
{
    unsigned int theUser(iNodeMode & LINUX_S_IRWXU);
    theUser >>= 3 * LINUX_S_IRWXU_OCTET;
    if (theUser >= 8)
    {
        theUser = 0;
    }
    return theUser;
}

unsigned int TTFSFileInfo::GetGroupPermissions() const
{
    unsigned int theGroup(iNodeMode & LINUX_S_IRWXG);
    theGroup >>= 3 * LINUX_S_IRWXG_OCTET;
    if (theGroup >= 8)
    {
        theGroup = 0;
    }
    return theGroup;
}

unsigned int TTFSFileInfo::GetOtherPermissions() const
{
    unsigned int theOthers(iNodeMode & LINUX_S_IRWXO);
    theOthers >>= 3 * LINUX_S_IRWXO_OCTET;
    if (theOthers >= 8)
    {
        theOthers = 0;
    }
    return theOthers;
}

unsigned int TTFSFileInfo::GetAccessTime() const
{
    return iINode.i_atime;
}

unsigned int TTFSFileInfo::GetChangeTime() const
{
    return iINode.i_ctime;
}

unsigned int TTFSFileInfo::GetModifiedTime() const
{
    return iINode.i_mtime;
}

unsigned int TTFSFileInfo::GetDeletionTime() const
{
    return iINode.i_dtime;
}

int lookup_entry(ext2_dir_entry* dirent, int offset, int blocksize, char* buf, void* private_data)
{
    Q_UNUSED(offset);
    Q_UNUSED(blocksize);
    Q_UNUSED(buf);
    ext2_dir_entry_2* dirent2 = (ext2_dir_entry_2*) dirent;
    TTFSLookupEntry *theEntry = static_cast<TTFSLookupEntry*>(private_data);
    if (theEntry->iEntryNameUtf8.size() != (dirent2->name_len & 0xFF))
    {
        return 0;
    }
    if (strncmp(theEntry->iEntryNameUtf8.data(), dirent2->name, (dirent2->name_len & 0xFF)))
    {
        return 0;
    }
    ext2_inode theInode;
    errcode_t theError = ext2fs_read_inode(theEntry->iFileSystem, dirent2->inode, &theInode);
    if (theError == 0)
    {
        if (dirent2->file_type == EXT2_FT_DIR)
        {
            theEntry->iNode = theInode;
            theEntry->iNode.i_mode |= LINUX_S_IFDIR;
            theEntry->iNodeNumber = dirent->inode;
            theEntry->iFound = true;
        }
        else if (dirent2->file_type == EXT2_FT_REG_FILE)
        {
            theEntry->iNode = theInode;
            theEntry->iNode.i_mode |= LINUX_S_IFREG;
            theEntry->iNodeNumber = dirent->inode;
            theEntry->iFound = true;
        }
        else 
        {
            ; // TODO
            // LINUX_S_IFSOCK
            // LINUX_S_IFLNK
            // LINUX_S_IFBLK
            // LINUX_S_IFCHR
            // LINUX_S_IFIFO
        }
    }
    if (theEntry->iFound)
    {
        return DIRENT_ABORT;
    }
    return 0;
}
