// TTFSFileInfo.h
//
// The class TTFSFileInfo represents file (or directory) information.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#pragma once

// Qt
#include <QtCore>

// TTFS
#include "TTFSTypes.h"

// ext2fs
#include <ext2fs_prefix.h>
#include <ext2fs/ext2fs.h>

class TTFSFile;
class TTFSDir;


class TTFSFileInfo
{
public:
    TTFSFileInfo();
    TTFSFileInfo(const QString& pathName);
    TTFSFileInfo(const TTFSFile& file);
    TTFSFileInfo(const TTFSDir& dir, const QString& fileName);
    TTFSFileInfo(const TTFSFileInfo& src);

    ~TTFSFileInfo();

    // Overriden virtual functions from QFileInfo.
    bool isFile() const;
    bool isDir() const;
    long long size() const;
    QString GetModeString() const;

    /** 
     * @brief
     * Test if the file path exists.
     *
     * @return
     * True, if the file path exists. Otherwise, false.
     */
    bool Exists() const;

    /**
     * @brief
     * Retrieve the owner permissions as unsigned integer: 0..7.
     */
    unsigned int GetOwnerPermissions() const;

    /**
     * @brief
     * Retrieve the user permissions as unsigned integer: 0..7.
     */
    unsigned int GetUserPermissions() const;

    /**
     * @brief
     * Retrieve the group permissions as unsigned integer: 0..7.
     */
    unsigned int GetGroupPermissions() const;

    /**
     * @brief
     * Retrieve the other permissions as unsigned integer: 0..7.
     */
    unsigned int GetOtherPermissions() const;

    unsigned int GetAccessTime() const;
    unsigned int GetChangeTime() const;
    unsigned int GetModifiedTime() const;
    unsigned int GetDeletionTime() const;

    const QString& GetAbsolutePath() const;
    const QString& GetFilename() const;
    QString GetPath() const;

public:
    TTFSFileInfo(const QString& aAbsolutePath, const QString& aFilename,
        TTFS::FileType aFileType, ext2_ino_t aNodePtr,
        ext2_inode aInode);

    TTFSFileInfo(const QString& aFilePath, TTFS::FileType aFileType,
                      ext2_ino_t aNodePtr, ext2_inode aInode);

    void SetParentDirectory(const QString& aParentDirectory);
    ext2_ino_t GetNodeNumber() const;

private:
    QString         iAbsolutePath;
    QString         iFilename;
    TTFS::FileType  iFileType;
    long long       iFileSize;
    ext2_ino_t      iNodePtr;
    __u16           iNodeMode;
    ext2_inode      iINode;
};

// Type TTFSFileInfoList is used for directory listings.
typedef QList<TTFSFileInfo> TTFSFileInfoList;
