// TTFSFileSystem.cpp
//
// Implementation of TTFSFileSystem class.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include <QDir>

#if defined(Q_WS_WIN)
#include "ext2WinPrefix.h"
#endif

#if defined(Q_WS_MAC)
#include "ext2MacPrefix.h"
#endif

#include <algorithm>
#include <ext2fs_prefix.h>
#include <ext2fs/ext2fs.h>

#include "TTFSFileSystem.h"
#include "TTFSNode.h"
#include "TTFSIO.h"

// Construct the static members with their default constructors.
std::auto_ptr<TTFSFileSystem> TTFSFileSystem::iInstance;

TTFSFileSystem::TTFSFileSystem()
{
    ; // empty
}

TTFSFileSystem::~TTFSFileSystem()
{
    for (TTFSNodeCollection::iterator theIterator(iCreatedTTFSNodes.begin());
    	theIterator != iCreatedTTFSNodes.end(); ++theIterator)
    {
        TTFSNodeList& theFileList = theIterator->second;
        for (TTFSNodeList::const_iterator theFileIterator = theFileList.begin();
             theFileIterator != theFileList.end(); ++theFileIterator)
        {
            TTFSNode* theNode = *theFileIterator;
            if (theNode)
            {
                theNode->unmount();
            }
        }  
        theFileList.clear();
    }
    iCreatedTTFSNodes.clear();
    
    // Remove the contents of the map.
    for (EXT3FileCollection::const_iterator theFileIter(iMountedEXT3Files.begin());
    	theFileIter != iMountedEXT3Files.end(); ++theFileIter)
    {
        if (theFileIter->second)
        {
            Close(theFileIter->second);
        }
    }
}

void TTFSFileSystem::Close(ext2_filsys aEx2Fs)
{
    errcode_t err = ext2fs_close(aEx2Fs);
    Q_ASSERT(err == 0);
}

/* static */ 
TTFSFileSystem& TTFSFileSystem::GetInstance()
{
    if (!iInstance.get())
    {
        Q_ASSERT((sizeof(__u64) >= 8) && 
            "ERROR: ext2fs library incorrectly compiled, sizeof(__u64) < 64 bits");
        iInstance.reset(new TTFSFileSystem());
    }
    return *iInstance;
}

/* static */
QString TTFSFileSystem::GetVolume(const QString& aPath)
{
    QString thePath;
    int thePos = aPath.indexOf(":");
    if (thePos > -1)
    {
        thePath = aPath.left(thePos + 1) + "/";
    }
    return thePath;
}

void TTFSFileSystem::Register(TTFSNode* aNodePtr)
{
    Q_ASSERT(aNodePtr);
    ext2_filsys theFS = aNodePtr->fs();
    TTFSNodeCollection::iterator theFileIterator = iCreatedTTFSNodes.find(theFS);
    if (iCreatedTTFSNodes.end() == theFileIterator)
    {
        iCreatedTTFSNodes.insert(std::make_pair(theFS, TTFSNodeList()));
    }
    iCreatedTTFSNodes[theFS].push_front(aNodePtr);
}

void TTFSFileSystem::Unregister(TTFSNode* aNodePtr)
{
    Q_ASSERT(aNodePtr);
    ext2_filsys theFS = aNodePtr->fs();
    TTFSNodeCollection::iterator theFileIterator = iCreatedTTFSNodes.find(theFS);
    if (iCreatedTTFSNodes.end() != theFileIterator)
    {
        theFileIterator->second.remove(aNodePtr);
        if (theFileIterator->second.size() == 0)
        {
            iCreatedTTFSNodes.erase(theFS);
        }
    }
}

bool TTFSFileSystem::IsReadOnly(const QString& aVolumeId) const
{
    bool theRetval(false);
    ext2_filsys theFileSystemPtr = GetRootEXT2(aVolumeId);
    if (theFileSystemPtr)
    {
        theRetval = !(theFileSystemPtr->flags & EXT2_FLAG_RW);
    }
    return theRetval;
}

ext2_filsys TTFSFileSystem::GetRootEXT2(const QString& aVolumePath) const
{
    ext2_filsys theRetval(0);
    QString theVolumePath = GetVolume(aVolumePath);
    if (!theVolumePath.isEmpty())
    {
        EXT3FileCollection::const_iterator theFileIter(iMountedEXT3Files.find(theVolumePath));
        if (theFileIter != iMountedEXT3Files.end())
        {
            theRetval = theFileIter->second;
        }
    }
    return theRetval;
}

void TTFSFileSystem::RemoveEXT3FS(ext2_filsys aFileSystemPtr)
{
    QString theImagePath(QString::fromAscii(aFileSystemPtr->device_name));
    QDir theDir(theImagePath);
    if (theDir.exists(theImagePath)) 
    {
        TTFSNodeCollection::iterator theFileIterator = iCreatedTTFSNodes.find(aFileSystemPtr);
        if (theFileIterator != iCreatedTTFSNodes.end())
        {
            TTFSNodeList& theFileList = theFileIterator->second;
            for (TTFSNodeList::const_iterator theFileIterator = theFileList.begin();
                 theFileIterator != theFileList.end(); ++theFileIterator)
            {
                TTFSNode* theNode = *theFileIterator;
                if (theNode)
                {
                    theNode->unmount();
                }
            }  
            theFileList.clear();
            iCreatedTTFSNodes.erase(aFileSystemPtr);
        }
        Close(aFileSystemPtr);
    }
}

class EXT3FileCollectionLookup :
    public std::unary_function<EXT3FileCollection::value_type, bool>
{
public:
    EXT3FileCollectionLookup(const QString& aFilename) : iFilename(aFilename)
    {}

    bool operator()(EXT3FileCollection::value_type aValue)
    {
        return QString::compare(iFilename,
                                QString::fromAscii(aValue.second->device_name),
                                Qt::CaseInsensitive) == 0;
    }
private:
    QString iFilename;
};

ext2_filsys TTFSFileSystem::MountEXT3(const QString& aPath,
                                      QString& aOutVolumeID,
                                      bool aReadOnly)
{
    ext2_filsys theRetval(0);
    aOutVolumeID.clear();

    Q_ASSERT(!aPath.isEmpty());
    if (!aPath.isEmpty())
    {
        EXT3FileCollectionLookup theFilenameLookup(aPath);
        EXT3FileCollection::iterator theFileIter = find_if(iMountedEXT3Files.begin(),
                                                           iMountedEXT3Files.end(),
                                                           theFilenameLookup);
        if (theFileIter != iMountedEXT3Files.end())
        {
            Q_ASSERT(QString::compare(QString::fromAscii(theFileIter->second->device_name),
                                      aPath,
                                      Qt::CaseInsensitive) == 0 &&
                     "Found FS object with a different name than aFileName.");
            // got it: should not create a new EXT2 FileSystem object.
            aOutVolumeID = theFileIter->first;
            theRetval = theFileIter->second;
        }
        else
        {
            // Create the filesystem pointer
            std::string theFileName(aPath.toAscii());
            io_manager theManager = image_io_manager;

            Q_ASSERT(theManager != 0);
            int theFlags = EXT2_FLAG_JOURNAL_DEV_OK;
            if (!aReadOnly)
            {
                theFlags |= EXT2_FLAG_RW;
            }

            // Volume could be created and could be removed. 
            // The number of volumes is limited to 0xff.
            const int theVolumeMax = 255;
            const QChar theFillChar = QLatin1Char('0');
            for (int theVolume = 0; theVolume < theVolumeMax; ++theVolume)
            {
                QString theVolumeID = QString("ext%1:/").arg(theVolume, 2, 16, theFillChar);  // Hex
                if (iMountedEXT3Files.find(aOutVolumeID) == iMountedEXT3Files.end())
                {
                    errcode_t err = ext2fs_open2(theFileName.c_str(), 0, theFlags, 
                                                 0, 0, theManager, &theRetval);
                    if (err == 0)
                    {
                        aOutVolumeID = theVolumeID;
                        iMountedEXT3Files.insert(std::make_pair(aOutVolumeID, theRetval));
                    }
                    break;
                }
            }
            if (aOutVolumeID.isEmpty())
            {
                theRetval = ext2_filsys(0);
            }
        }
    }
    return theRetval;
}

void TTFSFileSystem::UnmountEXT3(const QString& aVolumeName)
{
    EXT3FileCollection::iterator theMountIter = iMountedEXT3Files.find(aVolumeName);
    if (theMountIter != iMountedEXT3Files.end())
    {
        ext2_filsys theFileSystemPtr = theMountIter->second;
        if (theFileSystemPtr != 0)
        {
            RemoveEXT3FS(theFileSystemPtr);
        }

        iMountedEXT3Files.erase(theMountIter);
    }
}

QString TTFSFileSystem::GetDeviceName(QString const& aVolumeID) const
{
    QString theRetval;
    EXT3FileCollection::const_iterator theFileIter(iMountedEXT3Files.find(aVolumeID));
    if (theFileIter != iMountedEXT3Files.end())
    {
        ext2_filsys theFS = theFileIter->second;
        theRetval = QString::fromAscii(theFS->device_name);
    }
    return theRetval;
}

QStringList TTFSFileSystem::GetVolumes() const
{
    QStringList theVolumes;
    for (EXT3FileCollection::const_iterator theIterator(iMountedEXT3Files.begin());
    	theIterator != iMountedEXT3Files.end(); ++theIterator)
    {
        theVolumes.push_back(theIterator->first);
    }
    return theVolumes;
}
