// TTFSFileSystem.h
//
// Declares TTFSFileSystem class. 
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#pragma once

// TTFS
#include "TTFSTypes.h"

// Qt
#include <QtCore>
#include <QString>

// C and/or C++
#include <map>
#include <memory>

struct struct_ext2_filsys;
typedef struct struct_ext2_filsys *ext2_filsys;

class TTFSNode;

typedef std::list<TTFSNode*> TTFSNodeList;

// The collection of all mounted TTFS.
typedef std::map<QString, ext2_filsys>      EXT3FileCollection;
typedef std::map<ext2_filsys, TTFSNodeList> TTFSNodeCollection;

/**
 * @brief
 * The implementation of the TTFSFileSystem that is a Singleton (Meyer's).
 * This class manages the ext2fs volumes and determines volumes from files names.
 */
class TTFSFileSystem
{
public:
    /**
     * GetInstance()
     *
     * @return TTFSFileSystem 
     */
    static TTFSFileSystem& GetInstance();

    /**
     * @brief
     * Get the volume from the path.
     *
     * @return
     * The volume of the path.
     */
    static QString GetVolume(const QString& aPath);

    /**
     * @brief
     * Get the EXT2 file-system pointer for the given volume.
     *
     * @param aVolumePath
     * The volume (or the complete file path) to find a EXT2 FS pointer for.
     *
     * @return
     * NULL if the volume does not exist in collection of mounted images,
     * or a valid EXT2 file-system pointer.
     */
    ext2_filsys GetRootEXT2(const QString& aVolumePath) const;

    /**
     * @brief
     * Is the mounted image readonly?
     *
     * @param aVolumeId
     * The volume id is assigned to the image when it is mounted.
     *
     * @return true if the image mounted and readonly.
     */
    bool IsReadOnly(const QString& aVolumeId) const;

    /**
     * @brief
     * Mount an EXT2 image from the given file path.
     *
     * @param [in] aPath
     * The file system name to create a EXT2 FS pointer for.
     *
     * @param [out] aOutVolumeID
     * The VolumeID created for the given volume.
     *
     * @param [in] aReadOnly
     * The access mode indicating whether open a filesystem readonly or not.
     *
     * @return ext2_filsys, NULL if errors.
     *
     * @notes
     *  Supports maximum 255 file systems.
     */
    ext2_filsys MountEXT3(const QString& aPath,
                          QString& aOutVolumeID,
                          bool aReadOnly = false);

    /**
     * @brief
     * Remove EXT2 file system-pointer using volume name
     *
     * @param [in] aVolumeName
     * The volume name (full file path allowed).
     */
    void UnmountEXT3(const QString& aVolumeName);

    /**
     * @brief
     * Retrieve the device name for a given volume name.
     * The 'device name' is the name of the TTFS image file that is mounted.
     *
     * @param [in] aVolumeName
     * The volume name.
     *
     * @return
     * The device name.
     */
    QString GetDeviceName(QString const& aVolumeName) const;

    /**
     * @brief
     * Retrieve the list of mounted volumes.
     *
     * @return
     * Return the list of mounted volumes.
     */
    QStringList GetVolumes() const;

    /**
     * @brief 
     * Destructor to set the Singleton instance to NULL
     */
    virtual ~TTFSFileSystem();

    void Register(TTFSNode* aNodePtr);
    void Unregister(TTFSNode* aNodePtr);

private:
    TTFSFileSystem();                          ///< Disallow usage.

    /**
     * @brief
     * Close an EXT3 file system.
     *
     * @param [in] aFileSystemPtr
     * The EXT3 file system.
     */
    static void Close(ext2_filsys aFileSystemPtr);

    /**
     * @brief
     * Remove EXT3 file system.
     *
     * @param aFileSystemPtr
     * The file-system pointer to the file system within a collection.
     */
    void RemoveEXT3FS(ext2_filsys aFileSystemPtr);

private:
    EXT3FileCollection iMountedEXT3Files;               ///< The collection of mounted EXT3 images (file system).
    TTFSNodeCollection iCreatedTTFSNodes;               ///< The collection of opened file handles.

    static std::auto_ptr<TTFSFileSystem> iInstance;    ///< Singleton instance.
};
