// TTFSFileSystemManager.cpp
//
// Implementation of TTFSFileSystemManager class.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "TTFSFileSystemManager.h"

// TTFS (this software library)
#include "TTFSFileSystem.h"
#include "TTFSFileEngine.h"
#include "TTFSDebug.h"
#include "CopyFileThread.h"

// Qt
#include <QDir>
#include <QFile>
#include <QFileInfo>
#include <QFSFileEngine>


TTFSFileSystemManager::TTFSFileSystemManager(QObject* aParent) :
    QObject(aParent)
{
    ; // empty
}

TTFSFileSystemManager::~TTFSFileSystemManager()
{
    ; // empty
}

/*static*/
void TTFSFileSystemManager::SwitchOffSpyInfo()
{
    TTFS::SetSpyLevel(TTFS::spylevelOFF);
}

/*static*/
QString TTFSFileSystemManager::FindEXT3ImageInPath(const QString& aPath)
{
    QString theRetval;
    QString theFilePath(aPath);
    if (!TTFSFileEngine::canHandle(aPath)) // not EXT3 path
    {
        if (!theFilePath.endsWith(TTFS::ContainerFileName))
        {
            QFileInfo theFileInfo(aPath);
            if (theFileInfo.isDir())
            {
                QDir theDir(aPath);
                theFilePath = theDir.filePath(TTFS::ContainerFileName);
            }
        }
        if (QFile(theFilePath).exists())
        {
            theRetval = theFilePath;
        }
    }
    return theRetval;
}

/*static*/
QString TTFSFileSystemManager::MountEXT3(const QString& aPath, bool aReadOnly)
{
    QString theRetval;
    QString theImageFileName = FindEXT3ImageInPath(aPath);
    if (!theImageFileName.isEmpty())
    {
        TTFSFileSystem::GetInstance().MountEXT3(theImageFileName, theRetval, aReadOnly);
    }
    return theRetval;
}

/*static*/
QString TTFSFileSystemManager::UnmountEXT3(const QString& aPath)
{
    QString theRetval;
    bool theImageFound = IsPathEXT3(aPath);
    if (theImageFound)
    {
        QString theVolumeName = TTFSFileSystem::GetVolume(aPath);
        if (!theVolumeName.isEmpty())
        {
            TTFSFileSystem& theFileSystem = TTFSFileSystem::GetInstance();
            theRetval = theFileSystem.GetDeviceName(theVolumeName);
            theFileSystem.UnmountEXT3(theVolumeName);
        }
    }
    return theRetval;
}

void TTFSFileSystemManager::CopyFile(const QString& aSourceFile, const QString& aDestinationFolder,
                                  IProgressReporter* aReporter, IPackageValidator* aValidator)
{
    CopyFileThread* theThread = new CopyFileThread(aSourceFile, aDestinationFolder, aReporter, aValidator, this);
    connect(theThread, SIGNAL(finished()), this, SLOT(onCopyFileFinished()));
    theThread->start();
}

void TTFSFileSystemManager::onCopyFileFinished()
{
    CopyFileThread* theThread = static_cast<CopyFileThread*>(sender());
    CopyFileThread::Error theError = theThread->GetLastError();
    theThread->deleteLater();
    emit copyFileDone((int)theError);
}

bool TTFSFileSystemManager::CreateFolder(const QString& aNewFolderPath, QString& aErrorMessage)
{
    Q_UNUSED(aErrorMessage);
    QDir theDir(aNewFolderPath);
    return theDir.mkdir(aNewFolderPath);
}

bool TTFSFileSystemManager::DeletePath(const QString& aPath, QString& aErrorMessage, bool aRecursive)
{
    aErrorMessage.clear();
    QFileInfo theFileInfo(aPath);
    bool theRetval(true);
    if (theFileInfo.isFile())
    {
        QFile theFile(aPath);
        theRetval = theFile.remove();
    }
    else
    {
        QDir theDir(aPath);
        if (aRecursive)
        {
            // Remove files/dirs within directory first
            QStringList theList(theDir.entryList(QDir::AllEntries | QDir::Hidden | QDir::System));
            foreach (QString theEntry, theList)
            {
                if ((QString::compare(theEntry, ".") != 0) &&
                    (QString::compare(theEntry, "..") != 0))
                {
                    if (!DeletePath(theDir.filePath(theEntry), aErrorMessage, true))
                    {
                        theRetval = false;
                        break;
                    }
                }
            }
        }
        QString theDirName(theDir.dirName());
        theRetval = (theRetval && theDir.cdUp()) && theDir.rmdir(theDirName);
    }
    return theRetval;
}

/*static*/
bool TTFSFileSystemManager::IsPathEXT3(const QString& aPath)
{
    return TTFSFileEngine::canHandle(aPath);
}

/*static*/
bool TTFSFileSystemManager::IsEXT3ImageReadOnly(const QString& aPath)
{
    bool theRetval(false);
    QString theVolumePath = TTFSFileSystem::GetInstance().GetDeviceName(aPath);
    if (!theVolumePath.isEmpty())
    {
        theRetval = !QFileInfo(theVolumePath).isWritable();
    }
    return theRetval;
}

/*static*/
QStringList TTFSFileSystemManager::DriveList()
{
    QStringList theRetval;
    
    // iterate drive list.
    // Note. QDir::drives() returns the list provided by QFSFileEngine::drives().
    // There is no use for us; iterate two file engines.
    QFileInfoList theDriveList(QFSFileEngine::drives());
    foreach (QFileInfo theDriveInfo, theDriveList)
    {
        if (theDriveInfo.isRoot() &&
            theDriveInfo.isReadable() &&
            theDriveInfo.isWritable())
        {
            theRetval.push_back(theDriveInfo.path());
        }
    }
    const QStringList& theVolumes(TTFSFileSystem::GetInstance().GetVolumes());
    foreach (QString theVolume, theVolumes)
    {
        theRetval.push_back(theVolume);
    }
    return theRetval;
}

/*static*/
quint64 TTFSFileSystemManager::freeBytes(const QString& aPath)
{
    TTFSFileEngine theFileEngine(aPath);
    return theFileEngine.freeBytes();
}

/*static*/
bool TTFSFileSystemManager::SetPermissionsAll(const QString& aPath)
{
    TTFSFileEngine theFileEngine(aPath);
    return theFileEngine.setPermissions(QFile::ReadOwner |
        QFile::WriteOwner |
        QFile::ExeOwner |
        QFile::ReadGroup |
        QFile::WriteGroup |
        QFile::ExeGroup |
        QFile::ReadOther |
        QFile::WriteOther |
        QFile::ExeOther);
}

/*static*/
QString TTFSFileSystemManager::AddVolumeToPath(const QString& aVolumeId, const QString& aPath)
{
    QString theRetval(aPath);
    if (!theRetval.startsWith(aVolumeId))
    {
        QDir theDir(aVolumeId);
        // A leading forward slash messes with QDir::filePath()
        if (theRetval.startsWith("/"))
        { 
            theRetval = theRetval.mid(1);
        }
        theRetval = theDir.filePath(theRetval);
    }
    return theRetval;
}

