// TTFSFileSystemManager.h
//
// Declaration of the TTFSFileSystemManager class.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#pragma once

// TTFS
#include "TTFS_API.h"

// Qt
#include <QObject>
#include <QString>

class QByteArray;
class IProgressReporter;
class IPackageValidator;

class TTFS_API TTFSFileSystemManager : public QObject
{
    Q_OBJECT

public:
    TTFSFileSystemManager(QObject* aParent = 0);
    virtual ~TTFSFileSystemManager();

    /**
     * @brief
     * Switch off the spy info messages.
     */
    static void SwitchOffSpyInfo();

    /**
     * @brief
     * Is the given path point to a file or folder in an EXT3 image?
     *
     * @param aPath
     * The given path to a file or folder.
     *
     * @return
     * |true| if the path points to an entry in an EXT3 image.
     */
    static bool IsPathEXT3(const QString& aPath);

   /**
     * @brief
     * Find the EXT3 image in the path (directory).
     *
     * @return
     * The path name if found, else empty string.
     */
    static QString FindEXT3ImageInPath(const QString& aPath);
    
    /**
     * @brief
     * Check whether the path is located on a mounted image which is readonly.
     *
     * @return
     * |true| iff the path is on an EXT3 image which is readonly.
     */
    static bool IsEXT3ImageReadOnly(const QString& aPath);

    /**
     * @brief
     * Mount the EXT3 image file if it is in the path.
     *
     * @return
     * The volume of the new EXT3 image. Empty if unsuccessful.
     */
    static QString MountEXT3(const QString& aPath, bool aReadOnly = false);

    /**
     * @brief
     * Unmount the EXT3 image pointed to by the path.
     *
     * @return
     * The folder that contained the EXT3 image that was mounted.
     * Empty if the path is not an EXT3 path.
     */
    static QString UnmountEXT3(const QString& aPath);

    /**
     * @brief
     * Copy the source file to the destination folder.
     */
    void CopyFile(const QString& aSourceFile, const QString& aDestinationFolder,
                  IProgressReporter* aReporter = 0, IPackageValidator* aValidator = 0);

    /**
     * @brief
     * Create a new folder.
     *
     * @param [in] aNewFolder
     * The folder name to create.
     *
     * @param [out] aErrorMessage
     * The error message iff the operation was unsuccessful.
     *
     * @return
     * |true| iff the folder was successfully created.
     */
    bool CreateFolder(const QString& aNewFolder, QString& aErrorMessage);

    /**
     * @brief DeletePath
     * Delete the file/directory from the file system.
     *
     * @param aPath
     * Path to the file/directory to remove.
     *
     * @param aErrorMessage
     * Provide error message if the operation fails.
	 *
	 * @param aRecursive
	 * If true, removes the aPath recursively (if it is a directory).
     *
     * @return
     * |true| iff the path exists and operation is successful.
     */
    bool DeletePath(const QString& aPath, QString& aErrorMessage, bool aRecursive = false);

    static QStringList DriveList();

    static quint64 freeBytes(const QString& aPath);
    static bool SetPermissionsAll(const QString& aPath);
    static QString AddVolumeToPath(const QString& aVolumeId, const QString& aPath);

signals:
    void copyFileDone(int);

private slots:
    void onCopyFileFinished();
};
