// TTFSIO.cpp
//
// This is the image implementation of the I/O manager.
// It operates on an EXT2 file system that is spread over multiple image files.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "TTFSIO.h"

// Qt
#include <QFile>
#include <QVarLengthArray>

// ext2fs
#if defined(Q_WS_WIN)           // Compile for Microsoft Windows.
#include "ext2WinPrefix.h"
#endif

#if defined(Q_WS_MAC)           // Compile for MacOSX.
#include "ext2MacPrefix.h"
#endif

#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include <ext2fs/ext2_fs.h>
#include <ext2fs/ext2fs.h>

// C and/or C++
#include <stdio.h>
#include <cstdio>
#include <string.h>

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_ERRNO_H
#include <errno.h>
#endif

#include <fcntl.h>

extern "C"
{

/*
 * For checking structure magic numbers...
 */
#define EXT2_CHECK_MAGIC(struct, code) \
    if ((struct)->magic != (code)) return (code)

struct file_image
{
    ext2_off_t  size;
    QFile*      image;
};

struct image_private_data
{
    int                                 magic;
    QVarLengthArray<struct file_image>* files;
};

static errcode_t multi_files_image_open(const char* name, int flags, io_channel* channel);
static errcode_t multi_files_image_close(io_channel channel);
static errcode_t multi_files_image_set_blksize(io_channel channel, int blksize);
static errcode_t multi_files_image_read_blk(io_channel channel, unsigned long block, int count, void* data);
static errcode_t multi_files_image_write_blk(io_channel channel, unsigned long block, int count,const void* data);
static errcode_t multi_files_image_flush(io_channel channel);
static errcode_t multi_files_image_write_byte(io_channel channel, unsigned long offset, int size, const void* data);
static errcode_t multi_files_image_set_option(io_channel channel, const char* option, const char* arg);

static struct struct_io_manager struct_multi_files_image_manager =
{
    EXT2_ET_MAGIC_IO_MANAGER,
    "Multi Files Image I/O Manager",
    multi_files_image_open,
    multi_files_image_close,
    multi_files_image_set_blksize,
    multi_files_image_read_blk,
    multi_files_image_write_blk,
    multi_files_image_flush,
    multi_files_image_write_byte,
    multi_files_image_set_option,
    0,
    0,
    0,
    { 0 }
};

io_manager image_io_manager = 
     &struct_multi_files_image_manager;

const int cImageFilesMax = 256;

// File size constraint: the file size should be equal 2GB or less for last file.
const qint64 cFileSizeMax = (qint64)((qint64)1 << 31);

static errcode_t multi_files_image_open(const char *name, int flags,
                                        io_channel *channel)
{
    io_channel io                           = 0;
    struct image_private_data*  data        = 0;
    errcode_t                   retval      = 0;
    QIODevice::OpenMode         open_flags  = QIODevice::ReadOnly;

    struct file_image*  current             = 0;
    int                 i                   = 0;
    qint64              theFileSize         = 0;
    QFile*              image               = 0;
    char*               file_name           = 0;
    int                 file_name_length    = 0;
    QString             theMessage;

    // Expect a non-0 filename.
    if (!name || !name[0])
    {
        retval = EXT2_ET_BAD_DEVICE_NAME;
        goto cleanup;
    }

    // Create struct_io_channel.
    retval = ext2fs_get_mem(sizeof(struct struct_io_channel), &io);
    if (retval)
    {
        goto cleanup;
    }

    // Filename must end with character + 3 digits: X999.
    file_name_length = strlen(name);
    if ((file_name_length < 4) || 
        !isdigit(name[file_name_length - 1]) ||
        !isdigit(name[file_name_length - 2]) ||
        !isdigit(name[file_name_length - 3]) ||
        isdigit(name[file_name_length - 4]))
    {
        retval = -1;
        goto cleanup;
    }

    // Store filename.
    retval = ext2fs_get_mem(file_name_length + 1, &file_name);
    if (retval)
    {
        goto cleanup;
    }
    strncpy(file_name, name, file_name_length);
    file_name[file_name_length] = '\0';

    // Clear struct_io_channel.
    memset(io, 0, sizeof(struct struct_io_channel));
    io->magic = EXT2_ET_MAGIC_IO_CHANNEL;

    // Create image_private_data.
    retval = ext2fs_get_mem(sizeof(struct image_private_data), &data);
    if (retval)
    {
        goto cleanup;
    }
    memset(data, 0, sizeof(struct image_private_data));

    //  Initialize struct_io_channel.
    io->manager = image_io_manager;
    retval = ext2fs_get_mem(file_name_length + 1, &io->name);
    if (retval)
    {
        goto cleanup;
    }
    strncpy(io->name, name, file_name_length);
    io->name[file_name_length]  = '\0';
    io->private_data            = data;
    io->block_size              = 1024;
    io->read_error              = 0;
    io->write_error             = 0;
    io->refcount                = 1;

    data->files = new QVarLengthArray<struct file_image>;
    data->magic = EXT2_ET_MAGIC_UNIX_IO_CHANNEL;
    if (flags & IO_FLAG_RW)
    {
        open_flags |= QIODevice::ReadWrite;
    }
    else
    {
        open_flags |= QIODevice::ReadOnly;
    }

    // Determine file number from last 3 digits.
    i = QString(name).right(3).toInt();
    do
    {
        current = 0;
        image   = 0;

        // Check if image file exists.
        if (QFile::exists(file_name))
        {
            image = new QFile(file_name);
        }

        // File doesn't exist.
        if (!image)
        {
            // Were previous files opened?
            if (data->files->isEmpty())
            {
                // No, this was the first one.
                retval = errno;
                goto cleanup;
            }

            // Previous files were opened. 
            // This was the last one. 
            // Exit loop.
            break;
        }

        // Open image.
        if (!image->open(open_flags))
        {
            delete image;
            retval = errno;
            goto cleanup;
        }

        // Copy file image to current.
        retval = ext2fs_get_mem(sizeof(struct file_image), &current);
        if (retval)
        {
            image->close();
            delete image;
            retval = -1;
            goto cleanup;
        }
        memset(current, 0, sizeof(struct file_image));

#if defined(Q_WS_MAC)
        ::fcntl(image->handle(), F_NOCACHE, 1);
#endif
        theFileSize = image->size();
        if (!image->seek(theFileSize))
        {
            image->close();
            delete image;
            ext2fs_free_mem(&current);

            // Unsuccessful seek.
            retval = errno ? errno : EXT2_ET_LLSEEK_FAILED;
            goto cleanup;
        }

        // Check file size.
        if ((theFileSize < 1) || (theFileSize > cFileSizeMax))
        {   
            // File is bigger than |cFileSizeMax| - that is not acceptable.
            // Don't free image; it's referenced in the list already.
            ext2fs_free_mem(&current);
            retval = errno ? errno : EXT2_ET_LLSEEK_FAILED;
            goto cleanup;
        }

        // Store reference in list.
        current->image = image;
        current->size = (ext2_off_t)theFileSize;

        // Succesfully openend file, append file top chain.
        data->files->append(*current);
        ext2fs_free_mem(&current);
        current = 0;

        theMessage = QString("Read and mounted: %1 : %2").arg(QString::fromAscii(file_name)).arg(i);

        // Update filename to next file.
        ++i;
        sprintf(file_name + file_name_length - 3, "%03X", i);
    }
    while (i < cImageFilesMax);
    theMessage = QString("Successfully mounted %1 file(s).").arg(i - 1);

    if (file_name)
    {
        ext2fs_free_mem(&file_name);
    }
    *channel = io;
    return 0;

cleanup:
    if (data)
    {
        // Close all image files.
        for (int i = 0; i < data->files->size(); ++i)
        {
            if ((*data->files)[i].image)
            {
                (*data->files)[i].image->close();
                delete (*data->files)[i].image;
            }
        }
        delete data->files;
        ext2fs_free_mem(&data);
    }
    if (file_name)
    {
        ext2fs_free_mem(&file_name);
    }
    if (io)
    {
        ext2fs_free_mem(&io);
    }

    // Return some error.
    return (retval ? retval : -1);
}

static errcode_t multi_files_image_close(io_channel channel)
{
    struct image_private_data*  data    = 0;
    errcode_t                   retval  = 0;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data*) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    // Reduce reference count first.
    --channel->refcount;
    if (channel->refcount > 0)
    {
        return 0;
    }

    // Close all images.
    for (int i = 0; i < data->files->size(); ++i)
    {
        if ((*data->files)[i].image)
        {
            (*data->files)[i].image->close();
            delete (*data->files)[i].image;
        }
    }
    delete data->files;
    ext2fs_free_mem(&channel->private_data);
    if (channel->name)
    {
        ext2fs_free_mem(&channel->name);
    }
    ext2fs_free_mem(&channel);
    return retval;
}

static errcode_t multi_files_image_set_blksize(io_channel channel, int blksize)
{
    struct image_private_data* data = 0;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data*) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    channel->block_size = blksize;
    return 0;
}

static errcode_t multi_files_image_read_blk(io_channel channel, 
                                            unsigned long block, int count,
                                            void* buf)
{
    struct image_private_data*  data                = 0;
    struct file_image*          current             = 0;
    errcode_t                   retval              = 0;
    size_t                      size                = 0;
    int                         actual              = 0;
    int                         actual_part         = 0;
    size_t                      size_part           = 0;
    size_t                      size_pending        = 0;
    void*                       buf_part            = 0;
    unsigned int                theBlocksPerFile    = 0;
    unsigned int                theLocationInFile   = 0;
    int                         theFileIndex        = 0;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data*) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    // Block size of channel doesn't change in meanwhile, it set once.
    theBlocksPerFile    = (*data->files)[0].size / channel->block_size;
    theFileIndex        = block / theBlocksPerFile;
    theLocationInFile   = (block % theBlocksPerFile) * channel->block_size;

    size = ((count < 0) ? -count : (count * channel->block_size));
    actual = 0;
    if ((0 <= theFileIndex) && (theFileIndex < data->files->size()))
    {
        current = &(*data->files)[theFileIndex];
        if (current->image->pos() != theLocationInFile)
        {
            if (!current->image->seek(theLocationInFile))
            {
                retval = (errno ? errno : EXT2_ET_LLSEEK_FAILED);
                goto error_out;
            }
        }
        size_pending = size;
        buf_part = buf;
        for (int i = theFileIndex; size_pending && (i < data->files->size()); ++i)
        {
            current = &(*data->files)[i];

            size_part = size_pending;
            if (size_part > (current->size - theLocationInFile))
            {
                size_part = (current->size - theLocationInFile);
            }
            actual_part = current->image->read((char*) buf_part, size_part);
            if ((size_t)actual_part != size_part)
            {
                break;
            }
            buf_part            = ((unsigned char*)buf_part) + actual_part;
            actual              += actual_part;
            size_pending        -= actual_part;
            theLocationInFile   = 0;
        }
    }
    if ((size_t)actual != size)
    {
        if (actual < 0)
        {
            actual = 0;
        }
        retval = EXT2_ET_SHORT_READ;
        goto error_out;
    }
    return 0;

error_out:
    memset((char*) buf + actual, 0, size-actual);
    if (channel->read_error)
    {
        retval = (channel->read_error)(channel, block, count, buf,
                                       size, actual, retval);
    }
    return retval;

}

static errcode_t multi_files_image_write_blk(io_channel channel,
                                             unsigned long block, int count,
                                             const void* buf)
{
    size_t                          size                = 0;
    int                             actual              = 0;
    int                             actual_part         = 0;
    errcode_t                       retval              = 0;
    struct image_private_data*      data                = 0;
    struct file_image*              current             = 0;
    size_t                          size_part           = 0;
    size_t                          size_pending        = 0;
    const void*                     buf_part            = 0;
    unsigned int                    theBlocksPerFile    = 0;
    unsigned int                    theLocationInFile   = 0;
    int                             theFileIndex        = 0;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data*) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    // Block size of channel doesn't change in meanwhile, it set once.
    theBlocksPerFile = (*data->files)[0].size / channel->block_size;
    theFileIndex = block / theBlocksPerFile;
    theLocationInFile = (block % theBlocksPerFile) * channel->block_size;
    size = ((count < 0) ? -count : (count * channel->block_size));
    actual = 0;
    if ((0 <= theFileIndex) && (theFileIndex < data->files->size()))
    {
        current = &(*data->files)[theFileIndex];
        if (current->image->pos() != theLocationInFile)
        {    
            if (!current->image->seek(theLocationInFile))
            {
                retval = errno ? errno : EXT2_ET_LLSEEK_FAILED;
                goto error_out;
            }
        }
        size_pending = size;
        buf_part = buf;
        for (int i = theFileIndex; size_pending && (i < data->files->size()); ++i)
        {
            current = &(*data->files)[i];

            size_part = size_pending;
            if (size_part > (current->size - theLocationInFile))
            {
                size_part = (current->size - theLocationInFile);
            }
            actual_part = current->image->write((const char*) buf_part, size_part);
            if ((size_t)actual_part != size_part)
            {
                break;
            }
            buf_part            = ((unsigned char*) buf_part) + actual_part;
            actual              += actual_part;
            size_pending        -= actual_part;
            theLocationInFile   = 0;
        }
    }
    if ((size_t) actual != size)
    {
        retval = EXT2_ET_SHORT_WRITE;
        goto error_out;
    }
    return 0;

error_out:
    if (channel->write_error)
    {
        retval = (channel->write_error)(channel, block, count, buf,
                                        size, actual, retval);
    }

    return retval;
}

static errcode_t multi_files_image_flush(io_channel channel)
{
    struct image_private_data*  data    = 0;
    errcode_t                   retval  = 0;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data*) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    for (int i = 0; i < data->files->size(); ++i)
    {
        (*data->files)[i].image->flush();
    }
    return retval;
}

static errcode_t multi_files_image_write_byte(io_channel channel,
                                              unsigned long offset, int size,
                                              const void *buf)
{
    struct image_private_data*  data                = 0;
    size_t                      actual              = 0;
    size_t                      actual_part         = 0;
    errcode_t                   retval              = 0;
    struct file_image*          current             = 0;
    size_t                      size_part           = 0;
    size_t                      size_pending        = 0;
    const void*                 buf_part            = 0;
    unsigned int                theLocationInFile   = 0;
    int                         theFileIndex        = 0;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data*) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    // Block size of channel doesn't change in meanwhile, it set once.
    theFileIndex = (ext2_loff_t)offset / (ext2_loff_t)(*data->files)[0].size;
    theLocationInFile = (ext2_loff_t)offset % (ext2_loff_t)(*data->files)[0].size;

    actual = 0;
    if ((0 <= theFileIndex) && (theFileIndex < data->files->size()))
    {
        current = &(*data->files)[theFileIndex];
        if (current->image->pos() != theLocationInFile)
        {    
            if (!current->image->seek(theLocationInFile))
            {
                retval = errno ? errno : EXT2_ET_LLSEEK_FAILED;
                goto error_out;
            }
        }
        size_pending = size;
        buf_part = buf;
        for (int i = theFileIndex; size_pending && (i < data->files->size()); ++i)
        {
            current = &(*data->files)[i];
            size_part = size_pending;
            if (size_part > (current->size - theLocationInFile))
            {
                size_part = (current->size - theLocationInFile);
            }
            actual_part = current->image->write((const char* )buf_part, size_part);
            if ((size_t)actual_part != size_part)
            {
                break;
            }
            buf_part            = ((unsigned char*) buf_part) + actual_part;
            actual              += actual_part;
            size_pending        -= actual_part;
            theLocationInFile   = 0;
        }
    }
    if (actual != (size_t)size)
    {
        retval = EXT2_ET_SHORT_WRITE;
        goto error_out;
    }
    return 0;

error_out:
    if (channel->write_error)
    {
        retval = (channel->write_error)(channel, 0, 0, buf,
                                        size, actual, retval);
    }
    return retval;
}

static errcode_t multi_files_image_set_option(io_channel channel,
                                              const char *option,
                                              const char *arg)
{
    Q_UNUSED(channel);
    Q_UNUSED(option);
    Q_UNUSED(arg);
    return EXT2_ET_INVALID_ARGUMENT;
}

} /* extern "C" */
