// TTFSNode.cpp
//
// Implementation of TTFSNode class.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include "TTFSNode.h"

// TTFS (this software library)
#include "TTFSDebug.h"
#include "TTFSError.h"
#include "TTFSFileInfo.h"
#include "TTFSFileSystem.h"
#include "TTFSTypes.h"

// Qt
#include <QtCore>
#include <QString>
#include <QDir>

// ext2fs
#include <ext2fs_prefix.h>
#include <ext2fs/ext2fs.h>

// -----------------------------------------------------------------------------
// STATIC C CALLBACK FUNCTIONS
// -----------------------------------------------------------------------------

extern "C"
{

static int cbDirEmpty(ext2_dir_entry* aDirEnt, int aUnused1, int aUnused2, char* aUnused3, 
    void* aPrivateData)
{
    Q_UNUSED(aUnused1);
    Q_UNUSED(aUnused2);
    Q_UNUSED(aUnused3);
    Q_ASSERT(aDirEnt);
    spyINFO("aDirEnt:" << aDirEnt->name);
    int                 theRetval = DIRENT_ABORT; // abort iteration as soon as non-empty
    int*                theIsEmpty = (int*) aPrivateData;
    ext2_dir_entry_2*   aDirEnt2 = (ext2_dir_entry_2*) aDirEnt;

    if (aDirEnt2->inode == 0)                    // Root directory.
    {
        theRetval = 0;
    }
    else
    {
        if ((aDirEnt2->name_len == 1) &&         // "." directory.
            (aDirEnt->name[0] == '.'))
        {
            theRetval = 0;
        }
        else
        {
            if ((aDirEnt2->name_len == 2) &&     // ".." directory.
                (aDirEnt->name[0] == '.') &&
                (aDirEnt->name[1] == '.'))
            {
                theRetval = 0;
            }
        }
    }
    *theIsEmpty = theRetval == 0; // No need to include previous results. Interation aborted as soon as non-empty.
    return theRetval;
}

static int cbBlockFree(ext2_filsys fs, blk_t* blockNr, int unused1, void* unused2)
{
    Q_UNUSED(unused1);
    Q_UNUSED(unused2);
    spyINFO("fs:" << (fs ? 1 : 0) << "blockNr:" << *blockNr);
    Q_ASSERT(fs);
    Q_ASSERT(blockNr);

    blk_t block = *blockNr;
    ext2fs_block_alloc_stats(fs, block, -1);
    return 0;
}

} // extern "C"

// -----------------------------------------------------------------------------
// CTOR/DTOR
// -----------------------------------------------------------------------------

TTFSNode::TTFSNode(const QString& aFullPath, TTFS::FileType aFileType, TTFS::OpenMode aOpenMode) :
    iVolumeName(TTFSNode::volumeName(aFullPath)),
    iPathName(TTFSNode::pathName(aFullPath)),
    iFS(0),
    iOpenMode(aOpenMode),
    iNodeType(aFileType),
    iFlushNeeded(false),
    iError(TTFS::_errUNDEF)
{
    spyINFO("fullPath:" << aFullPath << "fileType:" << aFileType << "mode:" << aOpenMode);
    Q_ASSERT(!aFullPath.isEmpty());
    Q_ASSERT(iNodeType == TTFS::ftRegularFile ||
        iNodeType == TTFS::ftDirectory);

    bool isOk = false;
    errcode_t err = 0;

    // Check if this is a TTFS file system.
    iFS = TTFSFileSystem::GetInstance().GetRootEXT2(iVolumeName);
    if (iFS)
    {
        // Register this object with the file manager.
        TTFSFileSystem::GetInstance().Register(this);
        isOk = true;
    }
    else
    {
        // Volume is not recognized. 
        // No TTFS engine is available for this path.
        iError = TTFS::errNoTTFS;
    }

    // Clean up if something went wrong.
    if (isOk)
    {
        // All ok: 
        spyINFO("Created node object," << toString());

        Q_ASSERT(iFS);
        Q_ASSERT(!iError);
    }
    else
    {
        spyWARN("Cannot create node object," <<
            "errcode_t:" << err << toString());
    }
    Q_ASSERT(isValid());
}

TTFSNode::~TTFSNode()
{
    Q_ASSERT(isValid());

    unmount();

    // Unregister. Safe to unregister if 'this' was not registered yet.
    TTFSFileSystem::GetInstance().Unregister(this);
}

// -----------------------------------------------------------------------------
// PUBLIC MEMBER FUNCTIONS
// -----------------------------------------------------------------------------

QString TTFSNode::pathName() const
{
    Q_ASSERT(isValid());
    return iPathName;
}

QString TTFSNode::volumeName() const
{
    Q_ASSERT(isValid());
    return iVolumeName;
}

QString TTFSNode::fullPathName() const
{
    Q_ASSERT(isValid());
    return volumeName() + pathName();
}

QString TTFSNode::nodeName() const
{
    Q_ASSERT(isValid());
    return nodeName(iPathName);
}

bool TTFSNode::chown(__u16 uid, __u16 gid)
{
    spyINFO("uid:" << uid << "gid:" << gid << toString());
    iError = TTFS::_errUNDEF;
    bool isOk = false;

    // Check if volume exsist.
    if (iFS)
    {
        // Get the (most recent) inode.
        ext2_ino_t inode = inodeNr();
        if (inode)
        {
            iError = chown(iFS, inode, uid, gid);
            if (!iError)
            {
                isOk = true;
            }
        }
        else
        {
            iError = TTFS::errNodeDoesNotExist;
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSNode::exists() const
{
    Q_ASSERT(isValid());
    bool theRetval(false);
    // The node exists of the TTFS exists and the inode is fine.
    if (iFS)
    {
        theRetval = inodeNr() != 0;
    }
    return theRetval;
}

bool TTFSNode::hardLink(const QString& dest)
{
    spyINFO("dest:" << dest << toString());
    iError = TTFS::_errUNDEF;
    bool isOk = false;

    // Check if this is a TTFS file system.
    if (iFS)
    {
        iError = hardLink(iFS, iPathName, dest);
        isOk = iError == TTFS::_errUNDEF;
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSNode::setTime(time_t mod, time_t acc)
{
    spyINFO("mod:" << mod << "acc:" << acc << toString());
    iError = TTFS::_errUNDEF;
    bool isOk = false;

    // Check if this is a TTFS file system.
    if (iFS)
    {
        // Get the (most recent) inode.
        ext2_ino_t inode = inodeNr();
        if (inode)
        {
            iError = setTime(iFS, inode, mod, acc);
            isOk = !iError;
        }
        else
        {
            iError = TTFS::errNodeDoesNotExist; 
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSNode::readOnly() const
{
    spyINFO("node:" << fullPathName());
    iError = TTFS::_errUNDEF;
    bool isOk     = false;
    bool result   = true;

    // Check if this is a TTFS file system.
    if (iFS)
    {
        // Get the (most recent) inode.
        ext2_ino_t inode = inodeNr();
        if (inode)
        {
            iError = readOnly(iFS, inode, result);
            isOk = !iError;
        }
        else
        {
            iError = TTFS::errNodeDoesNotExist;
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));

    if (!isOk)
    {
        spyWARN(toString());
    }
    // Note that we return the read-only status, not the isOk flag.
    Q_ASSERT(isValid());
    return result;
}

bool TTFSNode::setReadOnly(bool readOnly)
{
    spyINFO("readOnly:" << readOnly << toString());
    iError = TTFS::_errUNDEF;
    bool isOk = false;

    // Check if this is a TTFS file system.
    if (iFS)
    {
        // Get the (most recent) inode.
        ext2_ino_t inode = inodeNr();
        if (inode)
        {
            iError = setReadOnly(iFS, inode, readOnly);
            isOk = !iError;
        }
        else
        {
            iError = TTFS::errNodeDoesNotExist;
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSNode::create()
{
    spyINFO(toString());
    iError = TTFS::_errUNDEF;
    bool isOk(false);

    // Check if this is a TTFS file system.
    if (iFS)
    {
        // Check if the file system was mounted as writable.
        if (iFS->flags & EXT2_FLAG_RW)
        {
            // Check if the ctor-supplied open mode allows writing.
            if (iOpenMode & TTFS::omWrite)
            {
                // Get the (most recent) inode.
                // Should NOT exist in this case!
                ext2_ino_t inode = inodeNr();
                if (!inode)
                {
                    // Try to create the new path (and get new inode to it).
                    iError = create(iFS, iPathName, true, isFile());
                    isOk = !iError;
                }
                else
                {
                    iError = TTFS::errNodeExists;
                }
            }
            else
            {
                iError = TTFS::errWritePermRequired;
            }
        }
        else
        {
            iError = TTFS::errFileSystemReadOnly;
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSNode::rename(const QString& dest)
{
    spyINFO("dest:" << dest<< toString());
    iError = TTFS::_errUNDEF;
    bool isOk = false;
    
    // Check if this is a TTFS file system.
    if (iFS)
    {
        // Get the (most recent) inode.
        ext2_ino_t inode = inodeNr();
        if (inode)
        {
            // Make sure the source/destination volume are identical.
            if (iVolumeName != volumeName(dest))
            {
                if (iPathName != pathName(dest))
                {
                    iError = rename(iFS, iPathName, pathName(dest));
                    if (!iError)
                    {
                        // All ok. Store new node name.
                        iPathName       = dest;
                        isOk            = true;
                    }
                }
                else
                {
                    // Source/destination name are identical. Do nothing.
                    isOk = true;
                }
            }
            else
            {
                iError = TTFS::errWrongDestVolume;
            }
        }
        else
        {
            iError = TTFS::errNodeDoesNotExist;
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}

bool TTFSNode::remove()
{
    spyINFO(toString());
    iError = TTFS::_errUNDEF;
    bool isOk = false;

    // Check if this is a TTFS file system.
    if (iFS)
    {
        // Supplied privilige must contain write rights.
        if (iOpenMode & TTFS::omWrite)
        {
            iError = remove(iFS, iPathName, isFile());
            isOk = (iError == 0);
        }
        else
        {
            iError = TTFS::errWritePermRequired;
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
    return isOk;
}

quint64 TTFSNode::usedBytes() const
{
    spyINFO("node:" << fullPathName());
    Q_ASSERT(isValid());
    iError = TTFS::_errUNDEF;
    bool    isOk    = false;
    quint64 amount  = 0;

    // Check if this is a TTFS file system.
    if (iFS)
    {
        if (iFS->super)
        {
            amount = iFS->super->s_blocks_count * EXT2_BLOCK_SIZE(iFS->super);
            isOk = true;
            spyINFO("amount:" << amount << toString());
        }
        else
        {
            iError = TTFS::errUnknownError;
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));

    // Note only the number of bytes is returned, not isOk.
    return amount;
}

quint64 TTFSNode::freeBytes() const
{
    spyINFO("node:" << fullPathName());
    Q_ASSERT(isValid());
    iError = TTFS::_errUNDEF;
    bool    isOk     = false;
    quint64 amount  = 0;

    // Check if this is a TTFS file system.
    if (iFS)
    {
        if (iFS->super)
        {
            amount = iFS->super->s_free_blocks_count * EXT2_BLOCK_SIZE(iFS->super);
            isOk = true;
            spyINFO("amount:" << amount << toString());
        }
        else
        {
            iError = TTFS::errUnknownError;
        }
    }
    else
    {
        iError = TTFS::errNoTTFS;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));

    // Note only the number of bytes is returned, not isOk.
    return amount;
}

TTFS::Error TTFSNode::lastError() const
{
    Q_ASSERT(isValid());
    Q_ASSERT(!iError || ((TTFS::_errMIN < iError) && (iError < TTFS::_errMAX)));
    return iError;
}

// -----------------------------------------------------------------------------
// PRIVATE HELPER FUNCTIONS
// -----------------------------------------------------------------------------

void TTFSNode::markDirty()
{
    spyINFO("node:" << fullPathName());
    errcode_t err = 0;
    if (iFS)
    {
        iFlushNeeded = true;
        err = ext2fs_read_bitmaps(iFS);
        if (!err)
        {
            ext2fs_mark_super_dirty(iFS);
            ext2fs_mark_ib_dirty(iFS);
            ext2fs_mark_bb_dirty(iFS);
        }
        else
        {
            spyWARN("ext2fs_read_bitmaps, errcode_t:" << err<< toString());
        }
    }
    else
    {
        spyWARN(toString());
    }
}

void TTFSNode::flush()
{
    spyINFO("node:" << fullPathName());

    // Check if this is a TTFS file system.
    if (iFS)
    {
        // Check if flush was required.
        if (iFlushNeeded)
        {
            // Flush the filesystem if needed.
            errcode_t err = ext2fs_flush(iFS);
            if (!err)
            {
                iFlushNeeded = false;
            }
            else
            {
                spyWARN("ext2fs_flush, errcode_t:" << err<< toString());
            }
        }
    }
    else
    {
        spyWARN(toString());
    }
}

void TTFSNode::unmount()
{
    spyINFO("node:" << fullPathName());
    iError = TTFS::_errUNDEF;
    bool isOk = false;

    // Check if this is a TTFS file system.
    if (iFS)
    {
        flush();
        isOk = true;
    }

    if (!isOk)
    {
        spyWARN(toString());
    }
    Q_ASSERT((isOk && !iError) || (!isOk && iError));
    Q_ASSERT(isValid());
}

ext2_ino_t TTFSNode::inodeNr() const
{
    // Do not issue spyINFO() message - too verbose.
    bool isOk = false;
    errcode_t err = 0;
    ext2_ino_t inode = 0;

    // Check if this is a TTFS file system.
    if (iFS)
    {
        // Get the inode of the path.
        err = ext2fs_namei(iFS, EXT2_ROOT_INO, EXT2_ROOT_INO,
            iPathName.toUtf8().data(), &inode);
        if (!err)
        {
            // The path was found, so the inode should be a good inode now.
            Q_ASSERT(inode);
            isOk = true;
        }
        else if (err == EXT2_ET_FILE_NOT_FOUND)
        {
            // The pathname was not found. This is not an error.
            // The file/dir simply does not exist and might be created
            // by the caller, later.
            Q_ASSERT(!inode);
            isOk = true;
        }
        else
        {
            ; // Something went wrong.
        }
    }

    if (!isOk)
    {
        spyWARN("errcode_t:" << err << toString());
    }
    Q_ASSERT(isValid());
    return inode;
}

ext2_filsys TTFSNode::fs() const
{
    return iFS;
}

// -----------------------------------------------------------------------------
// STATIC MEMBER FUNCTIONS
// -----------------------------------------------------------------------------

QString TTFSNode::pathName(const QString& aPath)
{
    // Strip volume part.
    QString theRetval = stripVolume(aPath);
    if (!theRetval.startsWith('/'))
    {
        theRetval.insert(0, '/');
    }
    Q_ASSERT(!theRetval.isEmpty());
    Q_ASSERT(theRetval.startsWith('/'));
    return theRetval;
}

QString TTFSNode::volumeName(const QString& aPath)
{
    // Volume defaults to "" (empty string).
    QString theVolume;

    // Check if volume name was found (ends with ':').
    int thePos = aPath.indexOf(':');
    if (thePos >= 0)
    {
        // Volume was found, return volume name up to ':'.
        theVolume = aPath.mid(0, thePos + 1);
    }
    Q_ASSERT(theVolume.isEmpty() || (theVolume.size() >= 2) && theVolume.endsWith(':'));
    return theVolume;
}

QString TTFSNode::nodeName(const QString& aPath)
{
    // Return the last part of the path name without the volume.
    QString theRetval = pathName(aPath);
    Q_ASSERT(theRetval.startsWith('/'));

    // Strip trailing '/' if it exists.
    if (theRetval.endsWith('/') && theRetval.size() != 1)
    {
        theRetval.chop(1);
    }
    Q_ASSERT(theRetval.startsWith('/'));

    // Get the part after the last '/'.
    int theIndex = theRetval.lastIndexOf('/');
    if (theIndex >= 0)
    {
        theRetval = theRetval.mid(theIndex + 1);
    }
    Q_ASSERT(!theRetval.contains('/'));
    return theRetval;
}

QString TTFSNode::parentFolder(const QString& aPath)
{
    QString theRetval = pathName(aPath);
    Q_ASSERT(theRetval.startsWith('/'));

    // Strip a trailing '/' if it exists.
    if (theRetval.endsWith('/') && theRetval.size() != 1)
    {
        theRetval.chop(1);
    }
    Q_ASSERT(theRetval.startsWith('/'));

    // Check if the path is the root directory.
    if (theRetval.size() != 1)
    {
        // Read until last '/'. There must be at least one '/'.
        int theIndex = theRetval.lastIndexOf('/');
        if (theIndex <= 0)
        {
            theIndex = 1; // leaves only the '/'
        }
        theRetval = theRetval.mid(0, theIndex);
    }
    Q_ASSERT(theRetval.startsWith('/') && (!theRetval.endsWith('/') ||  theRetval.size() == 1));
    return theRetval;
}

QString TTFSNode::stripVolume(const QString& aPath)
{
    QString theRetval(aPath);
    int theIndex = theRetval.indexOf(':');
    if (theIndex >= 0)
    {
        theRetval = theRetval.mid(theIndex + 1);
    }
    if (!theRetval.startsWith('/'))
    {
        theRetval.insert(0, '/');
    }
    Q_ASSERT(!theRetval.contains(':'));
    return theRetval;
}

bool TTFSNode::chmod(mode_t mode)
{
    spyINFO("mode:" << mode << toString());
    Q_ASSERT(isValid());
    Q_ASSERT((mode & LINUX_S_IFREG) || (mode & LINUX_S_IFDIR));

    iError = TTFS::_errUNDEF;
    bool isOk = false;
    errcode_t err = 0;

    // Read the inode.
    ext2_ino_t inode = inodeNr();
    ext2_inode  inodeStruct;
    err = ext2fs_read_inode(iFS, inode, &inodeStruct);
    if (!err)
    {
        // Set the correct file mode.
        inodeStruct.i_mode = mode;

        // Write the inode.
        err = ext2fs_write_inode(iFS, inode, &inodeStruct);
        if (err)
        {
            iError = TTFS::errWriteiNodeFailed;
            spyWARN("errcode_t:" << err << toString());
        }
        else
        {
            isOk = true;
        }
    }
    else
    {
        iError = TTFS::errReadiNodeFailed;
        spyWARN("errcode_t:" << err << toString());
    }
    return isOk;
}

TTFS::Error TTFSNode::chown(ext2_filsys fs, ext2_ino_t inode, __u16 uid, __u16 gid)
{
    spyINFO("fs:" << (fs ? 1 : 0) << "inode:" << inode << "uid:" << uid << "gid:" << gid);
    Q_ASSERT(fs);
    Q_ASSERT(inode);

    TTFS::Error errorCode = TTFS::_errUNDEF;
    errcode_t err = 0;

    ext2_inode  inodeStruct;
    err = ext2fs_read_inode(fs, inode, &inodeStruct);
    if (!err)
    {
        // Set the correct user and group ID.
        inodeStruct.i_gid = gid;
        inodeStruct.i_uid = uid;

        // Write the inode.
        err = ext2fs_write_inode(fs, inode, &inodeStruct);
        if (err)
        {
            spyWARN("ext2fs_write_inode, errcode_t:" << err << "inode:" << inode);
            errorCode = TTFS::errWriteiNodeFailed;
        }
    }
    else
    {
        spyWARN("ext2fs_read_inode, errcode_t:" << err << "inode:" << inode);
        errorCode = TTFS::errReadiNodeFailed;
    }

    if (errorCode)
    {
        spyWARN("TTFSError:" << errorCode << "errcode_t:" << err <<
            "inode:" << inode << "uid:" << uid << "gid:" << gid);
    }
    return errorCode;
}

TTFS::Error TTFSNode::create(ext2_filsys fs, const QString& dest, bool createPath, bool isFile)
{
    spyINFO("fs:" << (fs ? 1 : 0) << "dest:" << dest << "createPath:" << createPath);
    Q_ASSERT(fs);
    Q_ASSERT(!dest.isEmpty());
    Q_ASSERT(dest.startsWith('/'));

    TTFS::Error errorCode         = TTFS::_errUNDEF;
    errcode_t   err               = 0;
    bool        ifErrorUndoMark   = false;
    bool        ifErrorUndoWrite  = false;

    // Initialize new inode number and struct and set values to 0.
    ext2_ino_t newiNodeNr = 0;
    ext2_inode newiNodeStruct;
    memset(&newiNodeStruct, 0, sizeof(newiNodeStruct));

    // Create a list of all dirs up to dest. If dest is a directory itself, 
    // list contains dest as well.
    QString fileName;
    QStringList dirs;
    if (isFile)
    {
        QString theParentFolder(parentFolder(dest));
        dirs = theParentFolder.split('/');
        fileName = nodeName(dest);
    }
    else
    {
        dirs = dest.split('/');
    }

    ext2_ino_t parentiNodeNr = EXT2_ROOT_INO;
    err = ext2fs_read_bitmaps(fs);
    if (err)
    {
        goto cleanup;
    }

    foreach (QString dirName, dirs)
    {
        if (dirName.isEmpty())
        {
            continue;
        }

        QByteArray theDirUtf8(dirName.toUtf8());
        const char* theDirData = theDirUtf8.data();

            // Check if the directory exists.
        ext2_ino_t newDiriNodeNr = 0;
        err = ext2fs_namei_follow(fs, EXT2_ROOT_INO, parentiNodeNr,
            theDirData, &newDiriNodeNr);
        if (err && (err != EXT2_ET_FILE_NOT_FOUND))
        {
            spyWARN("ext2fs_namei_follow, errcode_t:" << err << "dest:" << dest <<
                "dirName:" << dirName << "parentiNodeNr:" << parentiNodeNr);
            goto cleanup;
        }

        // Check if the directory exists or not.
        if (err == EXT2_ET_FILE_NOT_FOUND)
        {
            Q_ASSERT(!newDiriNodeNr);

            // The directory does not exist.
            if (createPath)
            {
                err = ext2fs_mkdir(fs, parentiNodeNr, 0, theDirData);
                if (err)
                {
                    spyWARN("ext2fs_mkdir, errcode_t:" << err << "dest:" << dest <<
                        "dirName:" << dirName << "parentiNodeNr:" << parentiNodeNr);
                    errorCode = TTFS::errCreateFailed;
                    goto cleanup;
                }

                // Get the inode of the new dir.
                err = ext2fs_namei(fs, EXT2_ROOT_INO, parentiNodeNr,
                    theDirData, &newDiriNodeNr);
                if (err)
                {
                    spyWARN("ext2fs_namei, errcode_t:" << err << "dest:" << dest <<
                        "dirName:" << dirName << "parentiNodeNr:" << parentiNodeNr);
                    errorCode = TTFS::errReadiNodeFailed;
                    goto cleanup;
                }
            }
            else
            {
                // Folder does not exist and we are not allowed to create it.
                errorCode = TTFS::errNodeDoesNotExist;
                goto cleanup;
            }
        }

        // The directory already exsisted or is newly created.
        err = ext2fs_check_directory(fs, newDiriNodeNr);
        if (err)
        {
            spyWARN("ext2fs_checkdirectory, errcode_t:" << err << "dest:" << dest <<
                "newDiriNodeNr:" << newDiriNodeNr);
            goto cleanup;
        }
        parentiNodeNr = newDiriNodeNr;
    }

    // Now, check if we need to create a file as well.
    if (!fileName.isEmpty())
    {
        // Create an empty file. First, get an inode.
        __u16       mode        = (0644 | LINUX_S_IFREG);

        err = ext2fs_new_inode(fs, parentiNodeNr, mode, 0, &newiNodeNr);
        if (err)
        {
            spyWARN("ext2fs_new_inode, errcode_t:" << err << "dest:" << dest <<
                "parentiNodeNr:" << parentiNodeNr << "mode:" << mode);
            errorCode = TTFS::errCreateiNodeFailed;
            goto cleanup;
        }

        err = ext2fs_test_inode_bitmap(fs->inode_map, newiNodeNr);
        if (err)
        {
            spyWARN("ext2fs_test_inode_bitmap, errcode_t:" << err << "dest:" << dest <<
                    "newiNodeNr:" << newiNodeNr);
            errorCode = TTFS::errCreateiNodeFailed;
            goto cleanup;
        }

        // Need to update the inode statistics. 
        // It marks the inode bitmap and updates the free inodes count.
        ext2fs_inode_alloc_stats2(fs, newiNodeNr, +1, 0);
        
        ifErrorUndoMark = true;

        // Fill the inode struct.
        newiNodeStruct.i_mode          = mode;
        newiNodeStruct.i_atime         = 0;
        newiNodeStruct.i_size          = 0;
        newiNodeStruct.i_ctime         = newiNodeStruct.i_atime;
        newiNodeStruct.i_mtime         = newiNodeStruct.i_atime;
        newiNodeStruct.i_links_count   = 1;

        // Write the inode.
        // This function should be called when writing a new inode. It makes
        // sure that extra part of large inodes is initialized properly, 
        // creation time is set.        
        err = ext2fs_write_new_inode(fs, newiNodeNr, &newiNodeStruct);
        if (err)
        {
            spyWARN("ext2fs_write_new_inode, errcode_t:" << err << "dest:" << dest <<
                "newiNodeNr:" << newiNodeNr);
            errorCode = TTFS::errWriteiNodeFailed;
            goto cleanup;
        }
        ifErrorUndoWrite = true;
        
        QByteArray theFilenameUtf8(fileName.toUtf8());
        const char* theFilenameData = theFilenameUtf8.data();

        // Now, get a dirent. This may fail if there's no free space in the directory
        // for a new entry. In that case, expand the directory.
        err = ext2fs_link(fs, parentiNodeNr, theFilenameData,
            newiNodeNr, EXT2_FT_REG_FILE);
        if (err)
        {
            if (err == EXT2_ET_DIR_NO_SPACE)
            {
                err = ext2fs_expand_dir(fs, parentiNodeNr);
                if (err) 
                {
                    spyWARN("ext2fs_expand_dir, errcode_t:" << err << "dest:" << dest <<
                        "diriNodeNr:" << parentiNodeNr);
                    goto cleanup;
                }

                // Try linking again, after resize.
                err = ext2fs_link(fs, parentiNodeNr, theFilenameData,
                    newiNodeNr, EXT2_FT_REG_FILE);
                if (err)
                {
                    spyWARN("ext2fs_link, errcode_t:" << err << "dest:" << dest <<
                        "parentiNodeNr:" << parentiNodeNr << "fileName:" << fileName);
                    goto cleanup;
                }
            }
            else
            {
                spyWARN("ext2fs_link, errcode_t:" << err << "dest:" << dest <<
                    "parentiNodeNr:" << parentiNodeNr << "fileName:" << fileName);
                errorCode = TTFS::errLinkFailed;
                goto cleanup;
            }
        }
    }
    Q_ASSERT(!errorCode);
    return errorCode;

cleanup:
    // Undo writing the inode, if needed.
    if (ifErrorUndoWrite)
    {
        // Keep all err code for spy message.
        newiNodeStruct.i_links_count = 0;
        err = ext2fs_write_inode(fs, newiNodeNr, &newiNodeStruct);
        if (err)
        {
            spyWARN("ext2fs_write_inode, errcode_t:" << err << "dest:" << dest <<
                "newiNodeNr:" << newiNodeNr);
        }
    }

    // Mark the inode as unused, if needed.
    if (ifErrorUndoMark)
    {
        ext2fs_inode_alloc_stats2(fs, newiNodeNr, -1, 0);
    }

    // Make sure we have some error code, unknown if needed.
    if (!errorCode)
    {
        errorCode = TTFS::errUnknownError;
    }

    if (errorCode)
    {
        spyWARN("TTFSError:" << errorCode << "errcode_t:" << err << "dest:" << dest);
    }
    Q_ASSERT((TTFS::_errMIN < errorCode) && (errorCode < TTFS::_errMAX));
    return errorCode;
}

TTFS::Error TTFSNode::rename(ext2_filsys fs, const QString& src, const QString& dest)
{
    spyINFO("fs:" << (fs ? 1 : 0) << "src:" << src << "dest:" << dest);
    Q_ASSERT(fs);
    Q_ASSERT(src.startsWith('/'));
    Q_ASSERT(dest.startsWith('/'));

    TTFS::Error errorCode         = TTFS::_errUNDEF;
    errcode_t   err               = 0;
    bool        ifErrorUndoMark   = false;
    bool        ifErrorUndoWrite  = false;

    // Declare inode number and struct and fill with 0.
    ext2_ino_t newiNodeNr  = 0;
    ext2_inode newiNodeStruct;
    memset(&newiNodeStruct, 0, sizeof(newiNodeStruct));

    QString srcParent;
    QString destParent;
    QString srcName;
    QString destName;

    QByteArray theSrcParentUtf8;
    QByteArray theDestParentUtf8;
    QByteArray theSrcNameUtf8;
    QByteArray theDestNameUtf8;

    ext2_ino_t srcParentiNodeNr = 0;
    ext2_ino_t srcEntryiNodeNr = 0;
    ext2_inode srcEntryiNodeStruct;
    memset(&srcEntryiNodeStruct, 0, sizeof(srcEntryiNodeStruct));

    mode_t  mode = 0;
    int     flag = EXT2_FT_UNKNOWN;

    // Make sure the source/destination volume are identical.
    if (volumeName(src) != volumeName(dest))
    {
        errorCode = TTFS::errWrongDestVolume;
        goto cleanup;
    }

    // Create srcName and destName for the helper function.
    // Strip the trailing '/'.
    srcName = pathName(src);
    destName = pathName(dest);

    // Find parent directory of src, might be empty.
    srcParent = parentFolder(srcName);

    // Now, set srcName to node name only, no trailing '/'.
    srcName = nodeName(srcName);

    // Find parent directory of dest, might be empty.
    // Might be empty for /xxx (pos == 0).
    destParent = parentFolder(destName);

    // Now, set destName to node name only, no trailing '/'.
    destName = nodeName(destName);

    // Make sure the src and dest directory are the same.
    if (srcParent != destParent)
    {
        errorCode = TTFS::errDirsMustBeSame;
        goto cleanup;
    }
    Q_ASSERT(srcParent == destParent);

    err = ext2fs_read_bitmaps(fs);
    if (err)
    {
        goto cleanup;
    }

    theSrcParentUtf8 = srcParent.toUtf8();
    theDestParentUtf8 = destParent.toUtf8();
    theSrcNameUtf8 = srcName.toUtf8();
    theDestNameUtf8 = destName.toUtf8();

    // Get inode of the parent directory.
    err = ext2fs_namei_follow(fs, EXT2_ROOT_INO, EXT2_ROOT_INO,
        theSrcParentUtf8.data(), &srcParentiNodeNr);
    if (err)
    {
        errorCode = TTFS::errNameNotFound;
        goto cleanup;
    }

    // Check if the parent directory is indeed a directory.
    err = ext2fs_check_directory(fs, srcParentiNodeNr);
    if (err)
    {
        if (err == EXT2_ET_NO_DIRECTORY)
        {
            errorCode = TTFS::errMustBeDir;
        }
        else
        {
            spyWARN("ext2fs_check_directory, errcode_t:" << err <<
               "srcParent:" << srcParent << "srcParentiNodeNr:" << srcParentiNodeNr);
            ; // Some other error.
        }
        goto cleanup;
    }

    // Get inode of file within source directory.
    err = ext2fs_namei_follow(fs, EXT2_ROOT_INO, srcParentiNodeNr,
        theSrcNameUtf8.data(), &srcEntryiNodeNr);
    if (err)
    {
        errorCode = TTFS::errNodeDoesNotExist;
        goto cleanup;
    }

    // Read inode record of source.
    err = ext2fs_read_inode(fs, srcEntryiNodeNr, &srcEntryiNodeStruct);
    if (err)
    {
        spyWARN("ext2fs_read_inode, errcode_t:" << err << "srcName:" << srcName);
        errorCode = TTFS::errReadiNodeFailed;
        goto cleanup;
    }

    // Determine file type.
    mode = srcEntryiNodeStruct.i_mode;
    if (LINUX_S_ISREG(mode))
    {
        flag = EXT2_FT_REG_FILE;
    }
    else if (LINUX_S_ISDIR(mode))
    {
        flag = EXT2_FT_DIR;
    }
    else if (LINUX_S_ISLNK(mode))
    {
        flag = EXT2_FT_SYMLINK;
    }
    else
    {
        // Other node types are not supported.
        spyWARN("Unsupported mode:" << mode << "src:" << src);
        errorCode = TTFS::errUnsupportedOperation;
        goto cleanup;
    }

    // Create a new empty inode in the same directory.
    err = ext2fs_new_inode(fs, srcParentiNodeNr, mode, 0, &newiNodeNr);
    if (err)
    {
        spyWARN("ext2fs_new_inode, errcode_t:" << err <<
           "srcParent:" << srcParent << "srcParentiNodeNr:" << srcParentiNodeNr);
        errorCode = TTFS::errCreateiNodeFailed;
        goto cleanup;
    }

    // Mark the inode as used.
    err = ext2fs_mark_inode_bitmap(fs->inode_map, newiNodeNr);
    if (err)
    {
        spyWARN("ext2fs_mark_inode_bitmap, errcode_t:" << err << "newiNodeNr:" << newiNodeNr);
        errorCode = TTFS::errCreateiNodeFailed;
        goto cleanup;
    }
    ifErrorUndoMark = true;

    // Create an inode struct and fill it with the old values.
    memcpy(&newiNodeStruct, &srcEntryiNodeStruct, sizeof(newiNodeStruct));

    // Write the new inode struct: 
    // This function should be called when writing a new inode. It makes
    // sure that extra part of large inodes is initialized properly, 
    // creation time is set.
    err = ext2fs_write_new_inode(fs, newiNodeNr, &newiNodeStruct);
    if (err)
    {
        // Cleanup. Mark the inode as free again.
        spyWARN("ext2fs_write_new_inode, errcode_t:" << err << "newiNodeNr:" << newiNodeNr);
        errorCode = TTFS::errWriteiNodeFailed;
        goto cleanup;
    }
    ifErrorUndoWrite = true;
    
    // Creating the link may fail if there's no free space in the directory left.
    err = ext2fs_link(fs, srcParentiNodeNr, theDestNameUtf8.data(),
        newiNodeNr, flag);
    if (err)
    {
        // Check if the failure was lack of space in the directory.
        if (err == EXT2_ET_DIR_NO_SPACE)
        {
            // Resize dir and retry.
            err = ext2fs_expand_dir(fs, srcParentiNodeNr);
            if (err) 
            {
                spyWARN("ext2fs_expand_dir, errcode_t:" << err <<
                    "srcParent:" << srcParent << "srcParentiNodeNr:" << srcParentiNodeNr);
                goto cleanup;
            }
            else
            {
                // Try to link again.
                err = ext2fs_link(fs, srcParentiNodeNr, theDestNameUtf8.data(),
                    newiNodeNr, flag);
                if (err) 
                {
                    spyWARN("ext2fs_link, errcode_t:" << err <<
                        "srcParent:" << srcParent << "srcParentiNodeNr:" << srcParentiNodeNr <<
                        "destName:" << destName);
                    errorCode = TTFS::errLinkFailed;
                    goto cleanup;
                }
            }
        }
        else
        {
            errorCode = TTFS::errLinkFailed;
            goto cleanup;
        }
    }
    Q_ASSERT(!errorCode);
    return errorCode;

cleanup:
    // Undo writing the inode, if needed.
    if (ifErrorUndoWrite)
    {
        // Keep all err code for spy message.
        newiNodeStruct.i_links_count = 0;
        err = ext2fs_write_inode(fs, newiNodeNr, &newiNodeStruct);
        if (err)
        {
            spyWARN("ext2fs_write_inode, errcode_t:" << err << "newiNodeNr:" << newiNodeNr);
        }
    }

    // Mark the inode as unused, if needed.
    if (ifErrorUndoMark)
    {
        // Keep all err code for spy message.
        err = ext2fs_unmark_inode_bitmap(fs->inode_map, newiNodeNr);
        if (err)
        {
            spyWARN("ext2fs_unmark_inode_bitmap, errcode_t:" << err << "newiNodeNr:" << newiNodeNr);
        }
    }

    // Make sure we have some error code, unknown if needed.
    if (!errorCode)
    {
        errorCode = TTFS::errUnknownError;
    }

    if (errorCode)
    {
        spyWARN("TTFSError:" << errorCode << "errcode_t:" << err <<
            "src:" << src << "dest:" << dest);
    }
    Q_ASSERT((TTFS::_errMIN < errorCode) && (errorCode < TTFS::_errMAX));
    return errorCode;
}

TTFS::Error TTFSNode::remove(ext2_filsys fs, const QString& src, bool isFile)
{
    spyINFO("fs:" << (fs ? 1 : 0) << "src:" << src);
    Q_ASSERT(fs);
    Q_ASSERT(!src.isEmpty());

    TTFS::Error errorCode = TTFS::_errUNDEF;
    errcode_t   err = 0;
    ext2_ino_t  srciNodeNr = 0;
    ext2_ino_t  parentDiriNodeNr = 0;
    ext2_inode  newiNodeStruct;
    memset(&newiNodeStruct, 0, sizeof(newiNodeStruct));

    QString srcName;
    QString srcParentName;

    QByteArray theParentUtf8;
    QByteArray theFileUtf8;

    // Mounted volume must be writable.
    if (!(fs->flags & EXT2_FLAG_RW))
    {
        spyWARN("Mounted read-only, src:" << src);
        errorCode = TTFS::errFileSystemReadOnly;
        goto cleanup;
    }

    err = ext2fs_read_bitmaps(fs);
    if (err)
    {
        spyWARN("ext2fs_read_bitmaps, errcode_t:" << err);
        errorCode = TTFS::errReadBitmapsFailed;
        goto cleanup;
    }

    // Find the containing directory.
    srcParentName = parentFolder(src);
    srcName = nodeName(src);
    parentDiriNodeNr = EXT2_ROOT_INO;

    theParentUtf8 = srcParentName.toUtf8();
    theFileUtf8 = srcName.toUtf8();

    if (srcParentName != "/")
    {
        err = ext2fs_namei(fs, EXT2_ROOT_INO, EXT2_ROOT_INO,
            theParentUtf8.data(), &parentDiriNodeNr);
        if (err || !parentDiriNodeNr)
        {
            spyWARN("ext2fs_namei, errcode_t:" << err << "srcParentName:" << srcParentName);
            errorCode = TTFS::errNodeDoesNotExist;
            goto cleanup;
        }
    }

    // Don't rely on the existing iNodeNr, because it may already be
    // deleted or did not exist in the first place.
    srciNodeNr = 0;
    err = ext2fs_namei(fs, EXT2_ROOT_INO, parentDiriNodeNr,       //!! TODO namei or namei_follow?
        theFileUtf8.data(), &srciNodeNr);
    if (err || !srciNodeNr)
    {
        spyWARN("ext2fs_namei, errcode_t:" << err << "srcName:" << srcName);
        errorCode = TTFS::errNodeDoesNotExist;
        goto cleanup;
    }

    // Check if the file is a special file; this is unsupported.
    if (srciNodeNr < EXT2_FIRST_INODE(fs->super))
    {
        spyWARN("Unsupported operation for inode:" << srciNodeNr);
        errorCode = TTFS::errUnsupportedOperation;
        goto cleanup;
    }

    // Read the inode record.
    err = ext2fs_read_inode(fs, srciNodeNr, &newiNodeStruct);
    if (err)
    {
        spyWARN("ext2fs_read_inode, errcode_t:" << err << "src:" << src);
        errorCode = TTFS::errReadiNodeFailed;
        goto cleanup;
    }

    if (isFile)
    {
        if (LINUX_S_ISDIR(newiNodeStruct.i_mode))
        {
            spyWARN("Should be a file, src:" << src);
            errorCode = TTFS::errMustBeFile;
            goto cleanup;
        }
    }
    else
    {
        if (!LINUX_S_ISDIR(newiNodeStruct.i_mode))
        {
            spyWARN("Should be a dir, src:" << src);
            errorCode = TTFS::errMustBeDir;
            goto cleanup;
        }

        // Check if the directory is empty.
        int isEmpty = 0;
        err = ext2fs_dir_iterate(fs, srciNodeNr, 0, 0, &cbDirEmpty, &isEmpty);
        if (err)
        {
            spyWARN("ext2fs_dir_iterate, errcode_t:" << err << "srciNodeNr:" << srciNodeNr <<
                "src:" << src);
            errorCode = TTFS::errDirIterateFailed;
            goto cleanup;
        }

        // Directory must be empty to be removed.
        if (!isEmpty)
        {
            spyWARN("Directory not empty:" << src);
            errorCode = TTFS::errDirNotEmpty;
            goto cleanup;
        }
    }

    // Try to unlink the srciNodeNr from its parentDiriNodeNr.
    err = ext2fs_unlink(fs, parentDiriNodeNr, 0, srciNodeNr, 0);
    if (err)
    {
        spyWARN("ext2fs_unlink, errcode_t:" << err << "src:" << src <<
            "parentDiriNodeNr:" << parentDiriNodeNr << "srciNodeNr:" << srciNodeNr);
        errorCode = TTFS::errUnlinkFailed;
        goto cleanup;
    }

    Q_ASSERT(newiNodeStruct.i_links_count);
    if (isFile)
    {
        --newiNodeStruct.i_links_count;
    }
    else
    {
        // The ext2fs_unlink seems to change only the dirEnt, not the inode.
        // The FS is now inconsistent. Repair.
        newiNodeStruct.i_links_count  = 0;
    }
    newiNodeStruct.i_dtime = QDateTime::currentDateTime().toUTC().toTime_t();

    err = ext2fs_write_inode(fs, srciNodeNr, &newiNodeStruct);
    if (err)
    {
        spyWARN("ext2fs_write_inode, errcode_t:" << err << "src:" << src <<
            "srciNodeNr:" << srciNodeNr);
        errorCode = TTFS::errWriteiNodeFailed;
        goto cleanup;
    }

    // Clean up if no links left.
    if (newiNodeStruct.i_links_count == 0)
    {
        // Free inodes of srciNodeNr.
        err = ext2fs_block_iterate(fs, srciNodeNr, BLOCK_FLAG_DEPTH_TRAVERSE, 0, 
            cbBlockFree, 0);
        if (err)
        {
            spyWARN("ext2fs_block_iterate, errcode_t:" << err << "src:" << src <<
                "srciNodeNr:" << srciNodeNr);
            errorCode = TTFS::errBlockIterateFailed;
            goto cleanup;
        }

        ext2fs_inode_alloc_stats2(fs, srciNodeNr, -1, LINUX_S_ISDIR(newiNodeStruct.i_mode)); //!! TODO ISDIR or ISREG?
        ext2fs_unmark_inode_bitmap(fs->inode_map, srciNodeNr);
        if (err)
        {
            spyWARN("ext2fs_unmark_inode_bitmap, errcode_t:" << err << "src:" << src <<
                "srciNodeNr:" << srciNodeNr);
            errorCode = TTFS::errUnMapBlockFailed;
            goto cleanup;
        }
    }

    if (!isFile)
    {
        // Update parent inode data.
        ext2_inode parentiNodeStruct;
        err = ext2fs_read_inode(fs, parentDiriNodeNr, &parentiNodeStruct);
        if (err)
        {
            spyWARN("ext2fs_read_inode, errcode_t:" << err << "src:" << src <<
                "parentDiriNodeNr:" << parentDiriNodeNr);
            errorCode = TTFS::errReadiNodeFailed;
            goto cleanup;
        }

        // Parent should have at a link to the removed dir.
        if (parentiNodeStruct.i_links_count < 1)
        {
            spyWARN("Link count incorrect, i_links_count:" << parentiNodeStruct.i_links_count <<
                "src:" << src);
            errorCode = TTFS::errUnlinkFailed;
            goto cleanup;
        }

        // Decrease link count of parent and write inode info.
        Q_ASSERT(parentiNodeStruct.i_links_count > 0);
        --parentiNodeStruct.i_links_count;
        err = ext2fs_write_inode(fs, parentDiriNodeNr, &parentiNodeStruct);
        if (err)
        {
            spyWARN("ext2fs_read_inode, errcode_t:" << err << "src:" << src <<
                "parentDiriNodeNr:" << parentDiriNodeNr);
            errorCode = TTFS::errWriteiNodeFailed;
            goto cleanup;
        }
    }

    err = ext2fs_write_bitmaps(fs);
    if (err)
    {
        spyWARN("ext2fs_write_bitmaps, errcode_t:" << err << "src:" << src);
        errorCode = TTFS::errWriteBitmapsFailed;
        goto cleanup;
    }

    Q_ASSERT(!errorCode);
    return errorCode;

cleanup:
    spyWARN(src << "TTFSError:" << errorCode << "errcode_t:" << err);
    return errorCode;
}

TTFS::Error TTFSNode::hardLink(ext2_filsys fs, const QString& src, const QString& dest)
{
    spyINFO("fs:" << (fs ? 1 : 0) << "src:" << src << "dest:" << dest);
    Q_ASSERT(fs);
    Q_ASSERT(!src.isEmpty());
    Q_ASSERT(!dest.isEmpty());

    TTFS::Error errorCode = TTFS::_errUNDEF;
    errcode_t   err = 0;

    QString     srcName;
    QString     destName;
    QString     srcParentName;
    QString     destParentName;

    QByteArray theSrcParentUtf8;
    QByteArray theDestParentUtf8;
    QByteArray theSrcNameUtf8;
    QByteArray theDestNameUtf8;

    ext2_ino_t  srcParentiNodeNr = 0;
    ext2_ino_t  destParentiNodeNr = 0;
    ext2_ino_t  srcEntryiNodeNr = 0;

    ext2_inode srcEntryiNodeStruct;
    memset(&srcEntryiNodeStruct, 0, sizeof(srcEntryiNodeStruct));

    mode_t mode = 0;
    int flag = EXT2_FT_UNKNOWN;
    
    // The destination must reside on the same volume as the source.
    if (volumeName(src) != volumeName(dest))
    {
        spyWARN("Must be same volume, src:" << src << "dest:" << dest);
        errorCode = TTFS::errWrongDestVolume;
        goto cleanup;
    }

    // Remove volume from paths (and clean up paths).

    srcParentName = parentFolder(src);
    // Parent name is name up to '/' (not included).
    Q_ASSERT(!srcParentName.endsWith('/'));

    srcName = nodeName(src);
    // Now, set srcName to node name only, no trailing '/'.
    Q_ASSERT(!srcName.isEmpty());
    Q_ASSERT(!srcName.endsWith('/'));

    destName = nodeName(dest);
    // Now, set destName to node name only, no trailing '/'.
    Q_ASSERT(!destName.isEmpty());
    Q_ASSERT(!destName.endsWith('/'));

    destParentName = parentFolder(dest);
    // Parent name is name up to '/' (not included).
    Q_ASSERT(!destParentName.endsWith('/'));

    // Check if source/dest name is absolute and not just the root dir '/'.
    if ((srcParentName == "/") || (srcName.isEmpty()))
    {
        spyWARN("Cannot be root, src:" << src << "dest:" << dest);
        errorCode = TTFS::errCannotBeRoot;
        goto cleanup;
    }

    // The format of the source/destination is now:
    // /xxx, or /xxx/yyy, /xxx/yyy/zzz, ...
    // Make sure source and dest are not the same file.
    if (srcName == destName)
    {
        spyWARN("Must be different, src:" << src << "dest:" << dest);
        errorCode = TTFS::errCannotBeSame;
        goto cleanup;
    }

    theSrcParentUtf8 = srcParentName.toUtf8();
    theDestParentUtf8 = destParentName.toUtf8();
    theSrcNameUtf8 = srcName.toUtf8();
    theDestNameUtf8 = destName.toUtf8();

        // Get inode of the parent directory.
    srcParentiNodeNr = 0;
    err = ext2fs_namei_follow(fs, EXT2_ROOT_INO, EXT2_ROOT_INO,
        theSrcParentUtf8.data(), &srcParentiNodeNr);
    if (err)
    {
        spyWARN("ext2fs_namei_follow, errcode_t:" << err << "srcParentName:" << srcParentName);
        errorCode = TTFS::errNameNotFound;
        goto cleanup;
    }

    // Returns EXT2_ET_NO_DIRECTORY if srcParentiNodeNr is not a dir.
    err = ext2fs_check_directory(fs, srcParentiNodeNr);
    if (err)
    {
        if (err == EXT2_ET_NO_DIRECTORY)
        {
            errorCode = TTFS::errMustBeDir;
        }
        spyWARN("ext2fs_check_directory, errcode_t:" << err << "srcParentName:" << srcParentName);
        goto cleanup;
    }

    // Get inode of the dest parent directory.
    destParentiNodeNr = 0;
    err = ext2fs_namei_follow(fs, EXT2_ROOT_INO, EXT2_ROOT_INO,
        theDestParentUtf8.data(), &destParentiNodeNr);
    if (err)
    {
        spyWARN("ext2fs_namei_follow, errcode_t:" << err << "destParentName:" << destParentName);
        errorCode = TTFS::errNameNotFound;
        goto cleanup;
    }

    // Returns EXT2_ET_NO_DIRECTORY if destParentiNodeNr is not a dir.
    err = ext2fs_check_directory(fs, destParentiNodeNr);
    if (err)
    {
        if (err == EXT2_ET_NO_DIRECTORY)
        {
            errorCode = TTFS::errMustBeDir;
        }
        spyWARN("ext2fs_namei_follow, errcode_t:" << err << "destParentName:" << destParentName);
        goto cleanup;
    }

    // Get inode of file within source directory.
    srcEntryiNodeNr = 0;
    err = ext2fs_namei_follow(fs, EXT2_ROOT_INO, srcParentiNodeNr,
        theSrcNameUtf8.data(), &srcEntryiNodeNr);
    if (err)
    {
        spyWARN("ext2fs_namei_follow, errcode_t:" << err << "srcName:" << srcName);
        errorCode = TTFS::errNodeDoesNotExist;
        goto cleanup;
    }

    // Read inode of source.
    err = ext2fs_read_inode(fs, srcEntryiNodeNr, &srcEntryiNodeStruct);
    if (err)
    {
        spyWARN("ext2fs_read_inode, errcode_t:" << err << "srcName:" << srcName <<
            "srcEntryiNodeNr:" << srcEntryiNodeNr);
        errorCode = TTFS::errReadiNodeFailed;
        goto cleanup;
    }

    // Determine file type.
    mode = srcEntryiNodeStruct.i_mode;
    if (LINUX_S_ISREG(mode))
    {
        flag = EXT2_FT_REG_FILE;
    }
    else if (LINUX_S_ISDIR(mode))
    {
        flag = EXT2_FT_DIR;
    }
    else if (LINUX_S_ISLNK(mode))
    {
        flag = EXT2_FT_SYMLINK;
    }
    else
    {
        // Other node types are not supported.
        spyWARN("Unsupported file mode:" << mode << "src:" << src);
        errorCode = TTFS::errUnsupportedOperation;
        goto cleanup;
    }

    // Creating the link may fail if there's no free space in the directory left.
    err = ext2fs_link(fs, destParentiNodeNr, theDestNameUtf8.data(),
        srcEntryiNodeNr, flag);
    if (err)
    {
        // Check if the failure was lack of space in the directory.
        if (err == EXT2_ET_DIR_NO_SPACE)
        {
            // Resize the directory and retry once (perfectly valid situation).
            err = ext2fs_expand_dir(fs, destParentiNodeNr);
            if (err) 
            {
                spyWARN("ext2fs_expand_dir, errcode_t:" << err << "destParentName:" << destParentName);
                goto cleanup;
            }

            // Try linking again.
            err = ext2fs_link(fs, destParentiNodeNr, theDestNameUtf8.data(),
                srcEntryiNodeNr, flag);
            if (err) 
            {
                spyWARN("ext2fs_link, errcode_t:" << err << "destParentName:" << destParentName <<
                    "destName:" << destName);
                errorCode = TTFS::errLinkFailed;
                goto cleanup;
            }
        }
        else
        {
            spyWARN("ext2fs_link, errcode_t:" << err << "destParentName:" << destParentName <<
                "destName:" << destName);
            errorCode = TTFS::errLinkFailed;
            goto cleanup;
        }
    }
    Q_ASSERT(errorCode);
    return errorCode;

cleanup:
    // Simply returns false. Make sure we have an error status code.
    if (errorCode)
    {
        errorCode = TTFS::errUnknownError;
    }

    if (errorCode)
    {
        spyWARN("TTFSError:" << errorCode << "errcode_t:" << err <<
            "src:" << src << "dest:" << dest);
    }
    Q_ASSERT((TTFS::_errMIN < errorCode) && (errorCode < TTFS::_errMAX));
    return errorCode;
}

TTFS::Error TTFSNode::setTime(ext2_filsys fs, ext2_ino_t inode, time_t mod, time_t acc)
{
    spyINFO("fs:" << (fs ? 1 : 0) << "inode:" << inode << "mod:" << mod << "acc:" << acc);
    Q_ASSERT(fs);
    Q_ASSERT(inode);

    TTFS::Error errorCode = TTFS::_errUNDEF;
    errcode_t err = 0;

    // Read the inode.
    ext2_inode  inodeStruct;
    err = ext2fs_read_inode(fs, inode, &inodeStruct);
    if (!err)
    {
        // Update the time stamps in the inode.
        inodeStruct.i_mtime = mod;
        inodeStruct.i_atime = acc;

        // Write the inode.
        err = ext2fs_write_inode(fs, inode, &inodeStruct);
        if (err)
        {
            spyWARN("ext2fs_write_inode, errcode_t:" << err);
            errorCode = TTFS::errWriteiNodeFailed;
        }
    }
    else
    {
        spyWARN("ext2fs_read_inode, errcode_t:" << err);
        errorCode = TTFS::errReadiNodeFailed;
    }

    if (errorCode)
    {
        spyWARN("TTFSError:" << errorCode << "errcode_t:" << err << "inode:" << inode);
    }
    return errorCode;
}


// An absolute mode is an octal number constructed from the sum of one or
// more of the following values:
//
// 04000   (the set-user-ID-on-execution bit) Executable files with this bit
//         set will run with effective uid set to the uid of the file owner.
//         Directories with the set-user-id bit set will force all files and
//         sub-directories created in them to be owned by the directory owner
//         and not by the uid of the creating process, if the underlying file
//         system supports this feature: see chmod(2) and the suiddir option
//         to mount(8).
// 02000   (the set-group-ID-on-execution bit) Executable files with this bit
//         set will run with effective gid set to the gid of the file owner.
// 01000   (the sticky bit) See chmod(2) and sticky(8).
// 00400   Allow read by owner.
// 00200   Allow write by owner.
// 00100   For files, allow execution by owner.  For directories, allow the
//         owner to search in the directory.
// 00040   Allow read by group members.
// 00020   Allow write by group members.
// 00010   For files, allow execution by group members.  For directories,
//         allow group members to search in the directory.
// 00004   Allow read by others.
// 00002   Allow write by others.
// 00001   For files, allow execution by others.  For directories allow others
//         to search in the directory.
TTFS::Error TTFSNode::readOnly(ext2_filsys fs, ext2_ino_t inode, bool& readOnly)
{
    spyINFO("fs:" << (fs ? 1 : 0) << "inode:" << inode);
    Q_ASSERT(fs);
    Q_ASSERT(inode > 0);

    TTFS::Error errorCode = TTFS::_errUNDEF;
    errcode_t err = 0;

    readOnly = true;
    ext2_inode iNodeStruct;
    err = ext2fs_read_inode(fs, inode, &iNodeStruct);
    if (!err)
    {
        // Get file permission and check the user-writable bit.
        long mode  = iNodeStruct.i_mode;
        readOnly = !(mode & 0200);
    }
    else
    {
        spyWARN("ext2fs_read_inode, errcode_t:" << err);
        errorCode = TTFS::errReadiNodeFailed;
    }

    if (errorCode)
    {
        spyWARN("TTFSError:" << errorCode << "errcode_t:" << err);
    }
    return errorCode;
}

TTFS::Error TTFSNode::setReadOnly(ext2_filsys fs, ext2_ino_t inode, bool readOnly)
{
    spyINFO("fs:" << (fs ? 1 : 0) << "inode:" << inode << "readOnly:" << readOnly);
    Q_ASSERT(fs);
    Q_ASSERT(inode > 0);

    TTFS::Error errorCode = TTFS::_errUNDEF;
    errcode_t err = 0;

    ext2_inode iNodeStruct;
    err = ext2fs_read_inode(fs, inode, &iNodeStruct);
    if (!err)
    {
        // Get file permission.
        long mode  = iNodeStruct.i_mode;
        if (readOnly)
        {
            // Clear bit.
            mode = (mode & ~0200);  // User writable bit.
        }
        else
        {
            // Set bit.
            mode = (mode | 0200);   // User writable bit.
        }
        iNodeStruct.i_mode = mode;
        Q_ASSERT((readOnly && (mode & 0200)) || (!readOnly && !(mode & 0200)));
        
        // Write the inode.
        err = ext2fs_write_inode(fs, inode, &iNodeStruct);
        if (!err)
        {
            spyWARN("ext2fs_write_inode, errcode_t:" << err << "inode:" << inode);
            errorCode = TTFS::errWriteiNodeFailed;
        }
    }
    else
    {
        spyWARN("ext2fs_read_inode, errcode_t:" << err);
        errorCode = TTFS::errReadiNodeFailed;
    }

    if (errorCode)
    {
        spyWARN("TTFSError:" << errorCode << "errcode_t:" << err);
    }
    return errorCode;
}

QString TTFSNode::toString() const
{
    QString theRetval(QString("[path:%1,fs:%2,om:%3]")
                    .arg(fullPathName()).arg(iFS ? 1 : 0).arg(iOpenMode));
    if (iError != TTFS::_errUNDEF)
    {
        theRetval += QString(",error:%1").arg(TTFS::toString(iError));
    }
    return theRetval;
}

// -----------------------------------------------------------------------------
// DEBUG MEMBER FUNCTIONS
// -----------------------------------------------------------------------------

#ifndef QT_NO_DEBUG
bool TTFSNode::isValid() const
{
    bool isOk = true;
    Q_ASSERT(isOk = isOk && (iVolumeName.isEmpty() || (iVolumeName.endsWith(':'))));
    Q_ASSERT(isOk = isOk && iPathName.startsWith('/'));
    Q_ASSERT(isOk = isOk && ((TTFS::_omMIN < iOpenMode) && (iOpenMode < TTFS::_omMAX)));
    Q_ASSERT(isOk = isOk && (!iError || ((TTFS::_errMIN < iError) && (iError < TTFS::_errMAX))));
    return isOk;
}
#endif
