// TTFSNode.h
//
// TTFSNode provides functionality to handle file and directories within a
// TTFS file system.
//
// The MIT License (MIT)
// Copyright © 2012 TomTom International B.V.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
// to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of
// the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#pragma once

// TTFS
#include "TTFSError.h"
#include "TTFSTypes.h"

// Qt
#include <QtCore>

// ext2fs
#include <ext2fs_prefix.h>
#include <ext2fs/ext2fs.h>



class TTFSNode
{
public:
    // The ctor of TTFSNode is used to create an object to represent a file or 
    // directory within a TTFS file system.
    //
    // The 'pathName' parameter specifies the full of the file/directory. Files
    // must not have a trailing '/', directories do need to have a trailing '/'. 
    // The path may start with the volume identification, which has the format
    // 'extNN:/', or it must start with the root '/'.
    // Examples:
    //   "/", "/file", "/dir/", "/dir/file", "/dir/dir/", "/dir/dir/file",
    //   "ext00:/", "ext00:/file", "ext00:/dir/", "ext00:/dir/file"
    //
    // The 'mode' parameter specifies the requested priviliges for operating on the
    // specified file/directory.
    //
    TTFSNode(const QString& fullPath, TTFS::FileType fileType,
             TTFS::OpenMode openMode = TTFS::omRead);
    virtual ~TTFSNode();

    // Dir/file name.
    //
    // pathName()       Returns the absolute path of the node directory, within
    //                  the volume, starting with a '/' root slash.
    //                  For files, this is the the full file specification,
    //                  including file name, excluding the volume.
    //                  For directories, this would be the directory itself,
    //                  including the trailing '/', excluding the volume.
    //                  This value can appended to volumeName() to get a full
    //                  file/dir specification including the volume name.
    //                  Examples:
    //                    "/", "/file", "/dir/", "/dir/file", "/dir/dir/"
    //
    // volumeName()     Returns the volume name of the node. The volume name can be
    //                  empty, or it can be of the form "name:". No trailing '/'.
    //                  Examples:
    //                    "", "ext00:"
    //
    // fullPathName()   Returns the absolute path, including the volume ID,
    //                  of the object. This is the same as volumeName() + pathName().
    //
    // parentFolder()   Returns the absolute path of the node, without a trailing '/'.
    //                  If the node was the root directory of the volume itself,
    //                  an empty string is returned, however.
    //                  Examples:
    //                    "", "/", "/dir", "/dir/dir"
    //
    // nodeName()       Returns the name of the file or directory, without a path 
    //                  specification. The name does not contain any '/' (not even a
    //                  trailing '/' for directory names). The node name of the
    //                  root directory is empty.
    //                  Examples:
    //                    "", "dir", "file"
    //
    QString pathName() const;
    QString volumeName() const;
    QString fullPathName() const;
    QString parentFolder() const;
    QString nodeName() const;

    // Node information.
    //
    // isDir()          Returns true if the node object represents a directory.
    // 
    // isFile()         Returns true if the node object represents a file.
    //
    // size()           Returns the size (in bytes) of the file (or directory).
    //
    // mode()           Returns the full file mode bit set.
    // modeOwner()      Returns the file permission for owner, group and other.
    // modegroup()      These are extracted from the 'mode_t' value returned by
    // modeOther()      mode().
    //
    // timeAccessed()   Returns the last access time, change time, last modified
    // timeChanged()    time and the deletion time of a node.
    // timeModified()
    // timeDeleted()
    //
    virtual bool isDir() const  = 0;
    virtual bool isFile() const = 0;

    virtual quint64 size() const = 0;

    mode_t  mode() const;

    __u16   modeOwner() const;
    __u16   modeGroup() const;
    __u16   modeOther() const;

    __u32   timeAccessed() const;
    __u32   timeChanged() const;
    __u32   timeModified() const;
    __u32   timeDeleted() const;

    // Permissions.
    //
    // chmod()          Set the file permissions for a node. For the value of mode,
    //                  use LINUX_S_xxx mode flags from ext2fs.h.
    //
    // chown()          Set the user/group ID of a node. 
    //
    virtual bool chmod(mode_t mode);
    bool         chown(__u16 uid, __u16 gid);

    // Attributes.
    //
    // exists()         Returns true if the volume is recorgnized and the node
    //                  exists.
    //
    // hardLink()       Create a hard link from one node a (new) node. The new
    //                  node name must reside on the same volume. The source node
    //                  or the destination node cannot be the root directory.
    //                  The destination node name must be an absolute path, 
    //                  including the volume of the node (must match source node).
    //                  If the source node is a file, the destination node must be
    //                  a file name specification as well (no trailing). If the
    //                  source node was a directory, the destination node must
    //                  be a directory specification (trailing '/').
    //
    // setTime()        Set the modification and last access time for a node.
    //
    // readOnly()       Returns true if the node is not writable. Note that the
    //                  read-only status of the volume is not taken into account
    //                  in this function. This is only about the node atribute.
    //
    // setReadOnly()    Set the node to be read-only or writable, removing or setting
    //                  the user write permission from the node.
    //
    bool    exists() const;
    bool    hardLink(const QString& dest);
    bool    setTime(time_t mod, time_t acc);
    bool    readOnly() const;
    bool    setReadOnly(bool readOnly);

    // Creation.
    //
    // create()         Create a new file (with size 0) or directory. The node
    //                  is created using permission user-writable: 0644 (rw-r--r--)
    //                  If the node already exists, an error is returned.
    //                  The path to the node is automatically created as part of
    //                  the call.
    //
    // rename()         Rename the current node. The new destination name must reside
    //                  on the same volume and the path must be identical to the source
    //                  path, up to the containing directory fo the node.
    //
    // remove()         Remove the current node. If the node was a file and it was open,
    //                  the file is automatically closed.
    //
    bool            create();
    virtual bool    rename(const QString& dest);
    virtual bool    remove();

    // Volume information.
    //
    // usedBytes()      Returns the number of bytes used on the volume of this node.
    //
    // freeBytes()      Returns the number of available bytes on the volume of this
    //                  node.
    //
    quint64 usedBytes() const;
    quint64 freeBytes() const;

    // This class keeps a 'last error status' value in lastError() (like 'errno').
    // If a member function returns an error condition, additional information
    // may be obtained by calling status() before calling any other member function.
    //
    // Note that the value of errorStatus() is ONLY valid when an error occurred.
    // If a function returns successfully, the value of status NOT defined to be
    // any particular value. So, do not check for _errUNDEF to see if things are OK.
    //
    TTFS::Error lastError() const;          // Retrieve the status of the last function
                                            // that returned a bool.
    virtual void flush();                   // Flush changes to TTFS.

// Protected helper functions.
protected:
    void markDirty();                       // Mark file system dirty, needs to be flushed.

    friend class TTFSFileSystem;            // Grant access for TTFSFileSystem to 
                                            // call |onFileSystemUnmount()|. 

    ext2_ino_t inodeNr() const;             // Get (current) inode number.
    ext2_filsys fs() const;                 // Get (current) filsys.

    void unmount();

// Static helper functions.
public:
    static QString      pathName(const QString& path);
    static QString      nodeName(const QString& path);
    static QString      volumeName(const QString& path);
    static QString      parentFolder(const QString& path);
    static QString      stripVolume(const QString& path);

    // Helper functions to modify the ext2 file system.  Note that these helper functions
    // are guaranteed to return _errUNDEF if all was OK.
    static TTFS::Error    chown(ext2_filsys fs, ext2_ino_t inode, __u16 uid, __u16 gid);
    static TTFS::Error    create(ext2_filsys fs, const QString& dest, bool createPath, bool isFile);
    static TTFS::Error    rename(ext2_filsys fs, const QString& src, const QString& dest);
    static TTFS::Error    remove(ext2_filsys fs, const QString& src, bool isFile);
    static TTFS::Error    hardLink(ext2_filsys fs, const QString& src, const QString& dest);
    static TTFS::Error    setTime(ext2_filsys fs, ext2_ino_t inode, time_t mod, time_t acc);
    static TTFS::Error    readOnly(ext2_filsys fs, ext2_ino_t inode, bool& readOnly);
    static TTFS::Error    setReadOnly(ext2_filsys fs, ext2_ino_t inode, bool readOnly);

// Protected data members.
protected:
    QString                 iVolumeName;        // Volume name, no trailing '/'.
    QString                 iPathName;          // Absolute path on volume, starts/ends with '/'.
    ext2_filsys             iFS;                // TTFS engine, 0 if error.
    TTFS::OpenMode          iOpenMode;          // Open mode for operating on iPathName.
    const TTFS::FileType    iNodeType;          // Node type, file or directory.
    bool                    iFlushNeeded;       // Flush of file system required.

    mutable TTFS::Error iError;             // Exit status of last bool function.

// Unsupported operations.
private:
    TTFSNode();                                 // Unsupported.
    TTFSNode(const TTFSNode&);                  // Unsupported.
    const TTFSNode& operator=(const TTFSNode&); // Unsupported.

// Debug functions.
public:
    QString toString() const;
#ifndef QT_NO_DEBUG
    bool isValid() const;
#endif
};
